#!/usr/bin/env python3
#######################################################################

import logging
from pathlib import Path

from pmm_cfg_gen.utils.settings_utils_v1 import SettingsOutput, globalSettingsMgr
from pmm_cfg_gen.utils.plex_utils import PlexItemHelper

#######################################################################


def writeFile(fileName: str | Path, data: str):
    """
     Write data to file. If file doesn't exist it will be created. This is to avoid problems with non - existant directories
     
     @param fileName - Name of file to write
     @param data - Data to write to file ( string or file
    """
    logging.getLogger("pmm_cfg_gen").debug("Writing File: {}".format(fileName))

    p = Path(str(fileName))

    # Create a path to the parent directory if it doesn t exist.
    if not p.resolve().parent.exists():
        logging.getLogger("pmm_cfg_gen").debug(
            "Creating path: {}".format(p.resolve().parent)
        )
        p.resolve().parent.mkdir(parents=True, exist_ok=True)

    with open(p, "w") as f:
        f.write(data)


def formatLibraryItemPath(output: SettingsOutput, library=None, collection=None, item=None, pmm=None, librarySettings=None) -> Path:
    """
     Formats path with library and item information. This is used to make sure paths are formatted correctly when saving a library or item
     
     @param output - SettingsOutput object that contains settings for the currently edited file
     @param library - Library object that is going to be saved as a title
     @param item - Item object that is going to be saved as a title
     
     @return Path object that is ready to be saved to a file or None if there is no path to the
    """
    strPath = PlexItemHelper.formatString(output.pathFormat, library=library, collection=collection, item=item, pmm=pmm, librarySettings=librarySettings, cleanTitleStrings=True)

    return Path(output.path, strPath).resolve()



#!/usr/bin/env python3
###################################################################################################

import json
import logging
import os
from pathlib import Path

import jsonpickle
import requests

import urllib3
import urllib3.exceptions
from plexapi.library import LibrarySection
from plexapi.collection import Collection
from plexapi.video import Video
from plexapi.audio import Artist
from plexapi.server import PlexServer

from pmm_cfg_gen.utils.settings_utils_v1 import globalSettingsMgr, SettingsTemplateLibraryTypeEnum, SettingsTemplateFileFormatEnum, SettingsPlexLibrary
from pmm_cfg_gen.utils.file_utils import formatLibraryItemPath
from pmm_cfg_gen.utils.plex_stats import PlexStats
from pmm_cfg_gen.utils.plex_utils import PlexItemHelper, PlexVideoHelper, PlexCollectionHelper
from pmm_cfg_gen.utils.template_manager import TemplateManager
from pmm_cfg_gen.utils.template_filters import generateTpDbSearchUrl
from pmm_cfg_gen.utils.pmm_utils import PlexMetaManagerCache

###################################################################################################

# the following two lines disable InsecureRequestWarning
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


###################################################################################################
class PlexLibraryProcessor:
    _logger: logging.Logger

    __collectionProcessedCache: dict[str, list]
    __itemProcessedCache: dict[str, list]

    __plexMetaManagerCache: dict[str, PlexMetaManagerCache]

    __session: requests.Session

    __stats: PlexStats

    templateManager: TemplateManager

    plexServer: PlexServer
    plexLibrary: LibrarySection
    plexLibrarySettings: SettingsPlexLibrary

    pathLibrary: Path

    ###############################################################################################
    def __init__(self) -> None:
        self._logger = logging.getLogger("pmm_cfg_gen")
        self._displayHeader()

        self.__stats = PlexStats()

        self.__collectionProcessedCache = dict()
        self.__itemProcessedCache = dict()
        self.__plexMetaManagerCache = dict()

        self.templateManager = TemplateManager(
            globalSettingsMgr.settings.templates.getTemplateRootPath()
        )

    ###############################################################################################

    def process(self):
        self.__stats.timerProgram.start()

        self._connectToServer()

        if globalSettingsMgr.settings.plex.libraries is None:
            self._logger.debug("Loading Library list from plex.")
            globalSettingsMgr.settings.plex.libraries = sorted( 
                [
                    SettingsPlexLibrary(x.title)
                    for x in self.plexServer.library.sections()
                    if x.type == "movie" or x.type == "show"
                ], 
                key = lambda x: x.name
            )

        self._logger.info(
            "Libraries to process: {}".format(
                ",".join([ x.name for x in globalSettingsMgr.settings.plex.libraries ])
            )
        )
        for library in globalSettingsMgr.settings.plex.libraries:
            self._processLibrary(library)

        self.__stats.timerProgram.stop()
        self.__stats.calcTotals()

        self._displayStats()

    ###############################################################################################
    def _connectToServer(self):
        self._logger.info(
            "Connection to plex server: {}".format(
                globalSettingsMgr.settings.plex.serverUrl
            )
        )

        self.__session = requests.Session()
        self.__session.verify = False
        self.plexServer = PlexServer(
            globalSettingsMgr.settings.plex.serverUrl,
            globalSettingsMgr.settings.plex.token,
            session=self.__session,
        )

    def _loadLibrary(self, library: SettingsPlexLibrary) -> LibrarySection:
        self._logger.debug("Loading plex library: {}".format(library.name))

        self.plexLibrarySettings = library
        self.plexLibrary = self.plexServer.library.section(self.plexLibrarySettings.name)
        
        self.__stats.initLibrary(self.plexLibrarySettings.name)
        self.__collectionProcessedCache.update({self.plexLibrarySettings.name: list()})
        self.__itemProcessedCache.update({self.plexLibrarySettings.name: list()})
        self.__plexMetaManagerCache.update({self.plexLibrarySettings.name: PlexMetaManagerCache() })

        self.__stats.timerLibraries[self.plexLibrarySettings.name].start()

        self.pathLibrary = formatLibraryItemPath(
            globalSettingsMgr.settings.output, library=self.plexLibrary, librarySettings=self.plexLibrarySettings
        )

        self.pathLibrary.mkdir(parents=True, exist_ok=True)

        globalSettingsMgr.settings.runtime.currentWorkingPath = str(self.pathLibrary)
        globalSettingsMgr.settings.runtime.currentWorkingPathRelative = str(self.pathLibrary.relative_to(Path(globalSettingsMgr.settings.output.path).resolve()))

        self._logger.debug("Library Path: '{}'".format(self.pathLibrary))

        tplFiles = globalSettingsMgr.settings.templates.getTemplateByGroupAndLibraryType("library", self.plexLibrary.type)

        self._logger.debug(
            "Template Files for Library Type '{}': {}".format(self.plexLibrary.type, jsonpickle.dumps(tplFiles, unpicklable=False))
        )

        if tplFiles is not None:
            for tplFile in tplFiles:
                if globalSettingsMgr.settings.generate.isFormatEnabled(tplFile.format):
                    fileName = Path(self.pathLibrary, "{}.{}".format(self.plexLibrarySettings.path, tplFile.fileExtension))

                    self.templateManager.renderAndSave(
                        tplFile.fileName, fileName, {"library": self.plexLibrary}
                    )

        if globalSettingsMgr.settings.plexMetaManager.cacheExistingFiles:
            self._logger.debug("Checking for Plex Meta Manager Cache enablement for this library")
            if self.plexLibrarySettings.pmm_path is not None:
                self._logger.info("-" * 50)
                self._logger.info("Loading Plex Meta Manager File Cache")
                self._logger.debug("Plex Meta Manager Path: {}".format(self.plexLibrarySettings.pmm_path))
                self.__plexMetaManagerCache[self.plexLibrarySettings.name].processFolder(self.plexLibrarySettings.pmm_path) 
                self._logger.info("-" * 50)
                
        return self.plexLibrary

    def _processLibrary(self, library: SettingsPlexLibrary):
        self._logger.info("-" * 50)
        self._logger.info("Started Processing Library: '{}'".format(library.name))

        self._loadLibrary(library)

        self._logger.info("Processing Library Collections")
        collections = self.plexLibrary.collections()

        self.__stats.countsLibraries[self.plexLibrarySettings.name].collections.total = len(collections)
        self.__stats.countsLibraries[self.plexLibrarySettings.name].collections.processed = 0

        self._saveCollectionTemplates()

        for collection in collections:
            try:
                self._processCollection(collection.title, collection)
            except:
                self._logger.exception(
                    "Error Processing Collection: {}".format(collection.title)
                )

        self._logger.info("Processing Library Items")
        items = self.plexLibrary.all()

        self.__stats.countsLibraries[self.plexLibrarySettings.name].items.total = len(items)
        self.__stats.countsLibraries[self.plexLibrarySettings.name].items.processed = 0
        self.__stats.countsLibraries[self.plexLibrarySettings.name].calcTotals()

        for item in items:
            try:
                self._processMetadata(None, [item])
            except:
                self._logger.exception("Error Processing Item: {}".format(item.title))

        self.__stats.timerLibraries[self.plexLibrarySettings.name].stop()

        self.__stats.countsLibraries[self.plexLibrarySettings.name].calcTotals()

        self._logger.info("-" * 50)        
        self._sortCache()
        self._saveCollectionReport()
        self._saveItemReport()
        #self._saveReport()

    def _processCollection(self, itemTitle: str, item):
        self.__stats.countsLibraries[self.plexLibrarySettings.name].collections.processed += 1

        if PlexItemHelper.isPMMItem(item) or item.childCount == 0:
            self._logger.info(
                "[{}/{}] Skipping {}: '{}'. [Reason: Dynamic/Empty Collecton]".format(
                    self.__stats.countsLibraries[
                        self.plexLibrarySettings.name
                    ].collections.processed,
                    self.__stats.countsLibraries[
                        self.plexLibrarySettings.name
                    ].collections.total,
                    item.type,
                    item.title,
                )
            )
            return

        pmmItem = self.__plexMetaManagerCache[self.plexLibrarySettings.name].collectionItem_to_dict(item.title)

        if pmmItem is not None and self.plexLibrarySettings.pmm_delta is True:
            self.__stats.countsLibraries[self.plexLibrarySettings.name].collections.skipped += 1

            self._logger.info(
                "[{}/{}] Skipping {}: '{}'. [Reason: Plex Meta Manager Cache Hit. Delta only requested]".format(
                    self.__stats.countsLibraries[
                        self.plexLibrarySettings.name
                    ].collections.processed,
                    self.__stats.countsLibraries[
                        self.plexLibrarySettings.name
                    ].collections.total,
                    item.type,
                    item.title,
                )
            )
            
            return
        
        self._logger.info(
            "[{}/{}] Processing {}: '{}'".format(
                self.__stats.countsLibraries[
                    self.plexLibrarySettings.name
                ].collections.processed,
                self.__stats.countsLibraries[self.plexLibrarySettings.name].collections.total,
                item.type,
                itemTitle,
            )
        )

        self._addCollectionToProcessedCache(item, pmmItem)

        tplFiles = globalSettingsMgr.settings.templates.getTemplateByGroupAndLibraryType("collection", self.plexLibrary.type)
        if tplFiles is None:
            self._logger.warn("\tNo Collection Templates for type '{}' specifed".format(self.plexLibrary.type))

            return

        self._logger.debug(
            "\tTemplate Files for Collection Type '{}': {}".format(self.plexLibrary.type, jsonpickle.dumps(tplFiles, unpicklable=False))
        )

        fileNameBase = PlexItemHelper.formatString(globalSettingsMgr.settings.output.fileNameFormat.collections, library=self.plexLibrary, collection=item, item=None, pmm=pmmItem, cleanTitleStrings=True)

        self._logger.debug("\tBase FileName: {}".format(fileNameBase))

        for tplFile in tplFiles:
            try:
                if globalSettingsMgr.settings.generate.isFormatEnabled(tplFile.format) and globalSettingsMgr.settings.generate.isTypeEnabled(tplFile.type):

                    if tplFile.subFolder is not None:
                        fileName = Path(self.pathLibrary, "collections", tplFile.subFolder, "{}.{}".format(fileNameBase, tplFile.fileExtension))
                    else: 
                        fileName = Path(self.pathLibrary, "collections", "{}.{}".format(fileNameBase, tplFile.fileExtension))

                    if not os.path.exists(fileName) or globalSettingsMgr.settings.output.overwrite:
                        self.templateManager.renderAndSave(
                            tplFile.fileName, fileName, tplArgs={
                                "library": jsonpickle.dumps(self.plexLibrary, unpicklable=False),
                                "item": { 
                                    "metadata": item, 
                                    "pmm": pmmItem
                                }
                            } 
                        )
                    else:
                        self._logger.warn("\tCollection File Name '{}' Exists. Skipping...".format(fileNameBase))
                else:
                    self._logger.debug("\tGenerating format '{}' for Collections is not enabled. Skipping...".format(tplFile.format))
            except:
                self._logger.exception("\tError Processing Collection Template: {}".format(tplFile.fileName))

        childItems = item.items()
        if len(childItems) > 0:
            self.__stats.countsLibraries[self.plexLibrarySettings.name].items.total = len(
                childItems
            )
            self.__stats.countsLibraries[self.plexLibrarySettings.name].items.processed = 0

            self._processMetadata(collection=item, items=childItems)

    def _processMetadata(self, collection : Collection | None, items : list[Video]):

        tplFiles = globalSettingsMgr.settings.templates.getTemplateByGroupAndLibraryType("metadata", self.plexLibrary.type)
        if tplFiles is None:
            self._logger.warn("No Metadata Templates for type '{}' specifed".format(self.plexLibrary.type))

            return

        self._logger.debug(
            "Template Files for Metadata Type '{}': {}".format(self.plexLibrary.type, jsonpickle.dumps(tplFiles, unpicklable=False))
        )

        itemName: str = ""
        if collection is not None:
            itemName=collection.title
            
            pmmItem = self.__plexMetaManagerCache[self.plexLibrarySettings.name].collectionItem_to_dict(collection.title)
            
            fileNameBase = PlexItemHelper.formatString(globalSettingsMgr.settings.output.fileNameFormat.collections, library=self.plexLibrary, collection=collection, item=None, pmm=pmmItem, cleanTitleStrings=True)
        elif len(items) == 1 and (isinstance(items[0], Video) or isinstance(items[0], Artist)):
            itemName=items[0].title
            
            pmmItem = self.__plexMetaManagerCache[self.plexLibrarySettings.name].metadataItem_to_dict(items[0].title)
            
            fileNameBase = PlexItemHelper.formatString(globalSettingsMgr.settings.output.fileNameFormat.metadata, library=self.plexLibrary, collection=collection, item=items[0], pmm=pmmItem, cleanTitleStrings=True)
        else:
            self._logger.error("Invalid item attempted to be processed: {}".format(items))

            return
      
        self._logger.debug("Base FileName: {}".format(fileNameBase))

        itemsWithExtras: list[dict] = []
        
        for item in items:
            self.__stats.countsLibraries[self.plexLibrarySettings.name].items.processed += 1

            if pmmItem is not None and self.plexLibrarySettings.pmm_delta is True:
                self._logger.info(
                    "[{}/{}] Skipping {}: '{}'. [Reason: Plex Meta Manager Cache Hit. Delta only requested]".format(
                        self.__stats.countsLibraries[
                            self.plexLibrarySettings.name
                        ].items.processed,
                        self.__stats.countsLibraries[self.plexLibrarySettings.name].items.total,
                        item.type,
                        PlexItemHelper.formatItemTitle(item),
                    )
                )

                self.__stats.countsLibraries[self.plexLibrarySettings.name].items.skipped += 1

            elif PlexItemHelper.isPMMItem(item):
                self._logger.info(
                    "[{}/{}] Skipping {}: '{}'. [Reason: Dynamic item]".format(
                        self.__stats.countsLibraries[
                            self.plexLibrarySettings.name
                        ].items.processed,
                        self.__stats.countsLibraries[self.plexLibrarySettings.name].items.total,
                        item.type,
                        PlexItemHelper.formatItemTitle(item),
                    )
                )
            elif self._isItemProcessed(item):
                self._logger.info(
                    "[{}/{}] Skipping {}: '{}'. [Reason: Already Processed]".format(
                        self.__stats.countsLibraries[
                            self.plexLibrarySettings.name
                        ].items.processed,
                        self.__stats.countsLibraries[self.plexLibrarySettings.name].items.total,
                        item.type,
                        PlexItemHelper.formatItemTitle(item),
                    )
                )
            else:
                self._logger.info(
                    "[{}/{}] Processing {}: '{}'".format(
                        self.__stats.countsLibraries[
                            self.plexLibrarySettings.name
                        ].items.processed,
                        self.__stats.countsLibraries[self.plexLibrarySettings.name].items.total,
                        item.type,
                        PlexItemHelper.formatItemTitle(item),
                    )
                )

                pmmItem = self.__plexMetaManagerCache[self.plexLibrarySettings.name].metadataItem_to_dict(item.title, item.year if isinstance(item, Video) else None)

                self._addItemToProcessedCache(collection, item, pmmItem)

                itemDict = { "metadata": item, "pmm": pmmItem }

                if isinstance(item, Video):
                    seasons = []
                    if "childCount" in item.__dict__:
                        self._logger.debug("  Loading Seasons...")
                        seasons = item.seasons()

                        itemDict.update({"seasons": seasons})
                elif isinstance(item, Artist):
                    albums = item.albums()
                    itemDict.update({"albums": albums})

                    tracks = item.tracks()
                    itemDict.update({"tracks": tracks})

                itemsWithExtras.append(itemDict)

        # Do we have anything we need to process
        if len(itemsWithExtras) > 0:
            sorted(itemsWithExtras, key=lambda x: x["metadata"].year if "year" in x["metadata"].__dict__ else 0)

            for tplFile in tplFiles:
                try:
                    if globalSettingsMgr.settings.generate.isFormatEnabled(tplFile.format) and globalSettingsMgr.settings.generate.isTypeEnabled(tplFile.type):
                        
                        if tplFile.subFolder is not None:
                            fileName = Path(self.pathLibrary, "metadata", tplFile.subFolder, "{}.{}".format(fileNameBase, tplFile.fileExtension))
                        else: 
                            fileName = Path(self.pathLibrary, "metadata", "{}.{}".format(fileNameBase, tplFile.fileExtension))

                        if not os.path.exists(fileName) or globalSettingsMgr.settings.output.overwrite:
                            self.templateManager.renderAndSave(
                                tplFile.fileName, fileName, tplArgs={
                                    "library": jsonpickle.dumps(self.plexLibrary, unpicklable=False),
                                    "items": itemsWithExtras 
                                } 
                            )
                        else:
                            self._logger.warn("  Metadata File Name '{}.{}' Exists. Skipping...".format(fileNameBase, tplFile.fileExtension))
                    else:
                        self._logger.debug("  Generating format '{}' for Metadata is not enabled. Skipping...".format(tplFile.format))
                except:
                    self._logger.exception("Error Processing Metadata Template: {}".format(tplFile.fileName))
                    
    def _isCollectionProcessed(self, item) -> bool:
        it = next(
            (
                x
                for x in self.__collectionProcessedCache[self.plexLibrarySettings.name]
                if x["title"] == item.title
            ),
            None,
        )

        return it is not None

    def _addCollectionToProcessedCache(self, item, pmmItem):
        if self.plexLibrarySettings.name not in self.__collectionProcessedCache.keys():
            self.__collectionProcessedCache[self.plexLibrarySettings.name] = []

        if not self._isCollectionProcessed(item):
            tpdbEntry = {
                "title": item.title,
                "searchUrl": generateTpDbSearchUrl(item),
                "metadata": item,
                "pmm": pmmItem if pmmItem is not None else {},
            }

            self.__collectionProcessedCache[self.plexLibrarySettings.name].append(tpdbEntry)

    def _isItemProcessed(self, item) -> bool:
        it = next(
            (
                x
                for x in self.__itemProcessedCache[self.plexLibrarySettings.name]
                if x["title"] == PlexItemHelper.formatItemTitle(item)
            ),
            None,
        )

        return it is not None

    def _addItemToProcessedCache(self, collection, item, pmmItem):
        pi = PlexVideoHelper(item)

        if self.plexLibrarySettings.name not in self.__itemProcessedCache.keys():
            self.__itemProcessedCache[self.plexLibrarySettings.name] = []

        if not self._isItemProcessed(item):
            tpdbEntry = {
                "collection": collection.title if collection is not None else "",
                "title": PlexItemHelper.formatItemTitle(item),
                "searchUrl": generateTpDbSearchUrl(item),
                "ids": pi.guids,
                "metadata": item,
                "pmm": pmmItem if pmmItem is not None else {},
            }

            self.__itemProcessedCache[self.plexLibrarySettings.name].append(tpdbEntry)

    def _sortCache(self):
        self.__collectionProcessedCache[self.plexLibrarySettings.name] = sorted(
            self.__collectionProcessedCache[self.plexLibrarySettings.name],
            key=lambda x: x["title"],
        )
        
        self.__itemProcessedCache[self.plexLibrarySettings.name] = sorted(
            self.__itemProcessedCache[self.plexLibrarySettings.name], key=lambda x: "{}:{}".format(x["collection"], x["title"])
        )

    def _saveCollectionTemplates(self):
        if not globalSettingsMgr.settings.generate.isTypeEnabled("collection.template"):
            self._logger.debug("Skipping Collection Templates...")
            return

        self._logger.info("Saving Collection Templates...")

        tplFiles = globalSettingsMgr.settings.templates.getTemplateByGroupAndLibraryType("collection", SettingsTemplateLibraryTypeEnum.TEMPLATE)
        if tplFiles is None:
            self._logger.warn("No Collection Templates for type '{}' specifed".format(self.plexLibrary.type))

            return
                    
        self._logger.debug(
            "Template Files for Template Type '{}': {}".format(self.plexLibrary.type, jsonpickle.dumps(tplFiles, unpicklable=False))
        )

        fileNameBase = PlexItemHelper.formatString(globalSettingsMgr.settings.output.fileNameFormat.template, library=self.plexLibrary, collection=None, item=None, cleanTitleStrings=True)
        
        for tplFile in tplFiles:
            try:
                if globalSettingsMgr.settings.generate.isFormatEnabled(tplFile.format) and globalSettingsMgr.settings.generate.isTypeEnabled(tplFile.type):
                    
                    if tplFile.subFolder is not None:
                        fileName = Path(self.pathLibrary, "_templates", "{}.{}".format(fileNameBase, tplFile.fileExtension))
                    else: 
                        fileName = Path(self.pathLibrary, "_templates", "{}.{}".format(fileNameBase, tplFile.fileExtension))

                    if not os.path.exists(fileName) or globalSettingsMgr.settings.output.overwrite:
                        self.templateManager.renderAndSave(
                            tplFile.fileName, fileName, tplArgs={
                                                                "library": self.plexLibrary,
                                                                # "settings": globalSettingsMgr.settings
                                                            }
                        )
                    else:
                        self._logger.warn("  Template File Name '{}' Exists. Skipping...".format(fileNameBase))
                else:
                    self._logger.debug("  Generating format '{}' for Template is not enabled. Skipping...".format(tplFile.format))
            except:
                self._logger.exception("Failed generating collection template: '{}'".format(tplFile.fileName))

    def _saveCollectionReport(self):
        if not globalSettingsMgr.settings.generate.isTypeEnabled("report.any"):
            self._logger.debug("Skipping Saving Collection Report...")
            return

        self._logger.info("Saving Collection Report...")

        tplFiles = globalSettingsMgr.settings.templates.getTemplateByGroupAndLibraryType("collection", SettingsTemplateLibraryTypeEnum.REPORT)
        if tplFiles is None:
            self._logger.warn("No Collection Report Templates for type '{}' specifed".format(self.plexLibrary.type))

            return
                    
        self._logger.debug(
            "Template Files for Report Type '{}': {}".format(self.plexLibrary.type, jsonpickle.dumps(tplFiles, unpicklable=False))
        )

        fileNameBase = PlexItemHelper.formatString(globalSettingsMgr.settings.output.fileNameFormat.collectionsReport, library=self.plexLibrary, collection=None, item=None, cleanTitleStrings=True)
        
        for tplFile in tplFiles:
            try:
                if globalSettingsMgr.settings.generate.isFormatEnabled(tplFile.format) and globalSettingsMgr.settings.generate.isTypeEnabled(tplFile.type):
                    
                    if tplFile.subFolder is not None:
                        fileName = Path(self.pathLibrary, "reports", tplFile.subFolder, "{}.{}".format(fileNameBase, tplFile.fileExtension))
                    else: 
                        fileName = Path(self.pathLibrary, "reports", "{}.{}".format(fileNameBase, tplFile.fileExtension))

                    if not os.path.exists(fileName) or globalSettingsMgr.settings.output.overwrite:
                        self.templateManager.renderAndSave(
                            tplFile.fileName, fileName, tplArgs={
                                                                "library": self.plexLibrary,
                                                                "items": self.__collectionProcessedCache[self.plexLibrarySettings.name],
                                                                "stats": json.loads(
                                                                    str(
                                                                        jsonpickle.dumps(
                                                                            self.__stats.countsLibraries[self.plexLibrarySettings.name],
                                                                            unpicklable=False,
                                                                        )
                                                                    )
                                                                ),
                                                                "processingTime": self.__stats.timerLibraries[
                                                                    self.plexLibrarySettings.name
                                                                ].to_dict()
                                                            }
                        )
                    else:
                        self._logger.warn("  Report File Name '{}' Exists. Skipping...".format(fileNameBase))
                else:
                    self._logger.debug("  Generating format '{}' for Report is not enabled. Skipping...".format(tplFile.format))
            except:
                self._logger.exception("Failed generating collection report: '{}'".format(tplFile.fileName))

    def _saveItemReport(self):
        if not globalSettingsMgr.settings.generate.isTypeEnabled("report.any"):
            self._logger.debug("Skipping Saving Item Report...")
            return

        self._logger.info("Saving Item Report...")

        tplFiles = globalSettingsMgr.settings.templates.getTemplateByGroupAndLibraryType("metadata", SettingsTemplateLibraryTypeEnum.REPORT)
        if tplFiles is None:
            self._logger.warn("No Item Report Templates for type '{}' specifed".format(self.plexLibrary.type))

            return
                    
        self._logger.debug(
            "Template Files for Report Type '{}': {}".format(self.plexLibrary.type, jsonpickle.dumps(tplFiles, unpicklable=False))
        )

        fileNameBase = PlexItemHelper.formatString(globalSettingsMgr.settings.output.fileNameFormat.metadataReport, library=self.plexLibrary, collection=None, item=None, cleanTitleStrings=True)

        for tplFile in tplFiles:
            try:
                if globalSettingsMgr.settings.generate.isFormatEnabled(tplFile.format) and globalSettingsMgr.settings.generate.isTypeEnabled(tplFile.type):
                    
                    if tplFile.subFolder is not None:
                        fileName = Path(self.pathLibrary, "reports", tplFile.subFolder, "{}.{}".format(fileNameBase, tplFile.fileExtension))
                    else: 
                        fileName = Path(self.pathLibrary, "reports", "{}.{}".format(fileNameBase, tplFile.fileExtension))

                    if not os.path.exists(fileName) or globalSettingsMgr.settings.output.overwrite:
                        self.templateManager.renderAndSave(
                            tplFile.fileName, fileName, tplArgs={
                                                                "library": jsonpickle.dumps(self.plexLibrary, unpicklable=False),
                                                                "items": self.__itemProcessedCache[self.plexLibrarySettings.name],
                                                                "stats": json.loads(
                                                                    str(
                                                                        jsonpickle.dumps(
                                                                            self.__stats.countsLibraries[self.plexLibrarySettings.name],
                                                                            unpicklable=False,
                                                                        )
                                                                    )
                                                                ),
                                                                "processingTime": self.__stats.timerLibraries[
                                                                    self.plexLibrarySettings.name
                                                                ].to_dict(),                        
                                                            }
                        )
                    else:
                        self._logger.warn("  Report File Name '{}' Exists. Skipping...".format(fileNameBase))
                else:
                    self._logger.debug("  Generating format '{}' for Report is not enabled. Skipping...".format(tplFile.format))
            except:
                self._logger.exception("Failed generating item report: '{}'".format(tplFile.fileName))

    def _saveReport(self):
        if not globalSettingsMgr.settings.generate.isTypeEnabled("report.any"):
            self._logger.debug("Skipping Saving Report...")
            return

        self._logger.info("Saving Report...")

        tplFiles = globalSettingsMgr.settings.templates.getTemplateByGroupAndLibraryType("report", SettingsTemplateLibraryTypeEnum.REPORT)
        if tplFiles is None:
            self._logger.warn("No Report Templates defined")

            return
                    
        self._logger.debug(
            "Template Files for Report: {}".format(jsonpickle.dumps(tplFiles, unpicklable=False))
        )

        fileNameBase = PlexItemHelper.formatString(globalSettingsMgr.settings.output.fileNameFormat.report, library=self.plexLibrary, collection=None, item=None, cleanTitleStrings=True)

        for tplFile in tplFiles:
            try:
                if globalSettingsMgr.settings.generate.isFormatEnabled(tplFile.format) and globalSettingsMgr.settings.generate.isTypeEnabled(tplFile.type):
                    
                    if tplFile.subFolder is not None:
                        fileName = Path(self.pathLibrary, "reports", tplFile.subFolder, "{}.{}".format(fileNameBase, tplFile.fileExtension))
                    else: 
                        fileName = Path(self.pathLibrary, "reports", "{}.{}".format(fileNameBase, tplFile.fileExtension))

                    if not os.path.exists(fileName) or globalSettingsMgr.settings.output.overwrite:
                        self.templateManager.renderAndSave(
                            tplFile.fileName, fileName, tplArgs={
                                                                "library": jsonpickle.dumps(self.plexLibrary, unpicklable=False),
                                                                "collections": self.__collectionProcessedCache[self.plexLibrarySettings.name],
                                                                "items": self.__itemProcessedCache[self.plexLibrarySettings.name],
                                                                "stats": json.loads(
                                                                    str(
                                                                        jsonpickle.dumps(
                                                                            self.__stats.countsLibraries[self.plexLibrarySettings.name],
                                                                            unpicklable=False,
                                                                        )
                                                                    )
                                                                ),
                                                                "processingTime": self.__stats.timerLibraries[
                                                                    self.plexLibrarySettings.name
                                                                ].to_dict(),                        
                                                            }
                        )
                    else:
                        self._logger.warn("  Report File Name '{}' Exists. Skipping...".format(fileNameBase))
                else:
                    self._logger.debug("  Generating format '{}' for Report is not enabled. Skipping...".format(tplFile.format))
            except:
                self._logger.exception("Failed generating report: '{}'".format(tplFile.fileName))

    def _displayHeader(self):
        self._logger.info("-" * 50)
        self._logger.info("Please Meta Manager Configuration File Generator")
        self._logger.info("-" * 50)

    def _displayStats(self):
        # self._logger.debug(json.dumps(self.__timers, indent=4, unpicklable=False))

        self._logger.info("-" * 50)

        self._logger.info("Overall Statistics")

        self._logger.info(
            "  Total Processing Time: {}".format(
                self.__stats.timerProgram.elapsed_time_ts.to_str()
            )
        )
        self._logger.info(
            "  Collections Processed: {}".format(
                self.__stats.countsProgram.collections.total
            )
        )
        self._logger.info(
            "  Items Processed: {}".format(self.__stats.countsProgram.items.total - self.__stats.countsProgram.items.skipped)
        )

        self._logger.info(
            "  Items Skipped: {}".format(self.__stats.countsProgram.items.skipped)
        )

        for libraryName in self.__stats.timerLibraries.keys():
            try:
                libraryTimer = self.__stats.timerLibraries[libraryName]
                libaryCounts = self.__stats.countsLibraries[libraryName]

                self._logger.info("Statistics for Library: '{}'".format(libraryName))

                self._logger.info(
                    "  Processing Time: '{}'. Total Time: {}".format(
                        libraryName, libraryTimer.elapsed_time_ts.to_str()
                    )
                )

                self._logger.info(
                    "  Collections Processed: {}".format(libaryCounts.collections.total)
                )

                self._logger.info(
                    "  Collections Skipped: {}".format(libaryCounts.collections.skipped)
                )

                self._logger.info("  Items Processed: {}".format(libaryCounts.items.total - libaryCounts.items.skipped))
                self._logger.info("  Items Skipped: {}".format(libaryCounts.items.skipped))
            except:
                self._logger.exception("Failed displaying stats for library: '{}'".format(libraryName))
                
        self._logger.info("-" * 50)

    # def _unlockAllLibraryFields(self):

    #     #self._logger.debug("Possible Fields: {}".format(self.plexLibrary.listFields(self.plexLibrary.type)))

    #     for field in self.plexLibrary.listFields(self.plexLibrary.type):
    #         self._logger.info("Unlocking Field: '{}'".format(field))
    #         try:
    #             self.plexLibrary.unlockAllField(field)
    #         except:
    #             self._logger.exception("Failed unlocking field: '{}'".format(field))
        


#!/usr/bin/env python3
#######################################################################

import datetime

#######################################################################


class timespan:
    __maxWeeks: int = 99999999999

    __totalTimeInSeconds: float

    __weeks: int
    __days: int
    __hours: int
    __minutes: int
    __seconds: float

    def __init__(
        self,
        weeks: int | None = None,
        days: int | None = None,
        hours: int | None = None,
        minutes: int | None = None,
        seconds: float | None = None,
        dt: datetime.datetime | None = None,
        time: float | None = None,
        timedelta: datetime.timedelta | None = None,
    ) -> None:
        self.__totalTimeInSeconds = 0
        self.__weeks = 0
        self.__days = 0
        self.__hours = 0
        self.__minutes = 0
        self.__seconds = 0

        if weeks:
            self.add_weeks(weeks)
        if days:
            self.add_days(days)
        if hours:
            self.add_hours(hours)
        if minutes:
            self.add_minutes(minutes)
        if seconds:
            self.add_seconds(seconds)
        if dt:
            self.add_datetime(dt)
        if time:
            self.add_time(time)

    def add_weeks(self, weeks: int):
        self.add_days(weeks * 7)

        return self

    def add_days(self, days: int):
        self.add_hours(days * 24)

        return self

    def add_hours(self, hours: int):
        self.add_minutes(hours * 60)

        return self

    def add_minutes(self, minutes: int):
        self.add_seconds(minutes * 60)

        return self

    def add_seconds(self, seconds: float):
        self.__totalTimeInSeconds += seconds

        self.__refresh()

        return self

    def add_datetime(self, dt: datetime.datetime):
        self.add_seconds((dt - datetime.datetime(1970, 1, 1)).total_seconds())

        return self

    def add_time(self, time: float):
        self.add_seconds(time)

        return self

    def to_dict(self):
        return {
            "weeks": self.__weeks,
            "days": self.__days,
            "hours": self.__hours,
            "minutes": self.__minutes,
            "seconds": self.__seconds,
            "totalTimeInSeconds": self.__totalTimeInSeconds,
        }

    def to_str(self):
        return ", ".join(
            "%d %s" % (num, unit)
            for num, unit in zip(
                [
                    (self.total_seconds // d) % m
                    for d, m in (
                        (604800, self.__maxWeeks),
                        (86400, 7),
                        (3600, 24),
                        (60, 60),
                        (1, 60),
                    )
                ],
                ["wk", "d", "hr", "min", "sec"],
            )
            if num
        )

    def __refresh(self):
        result = [
            (num, unit)
            for num, unit in zip(
                [
                    (self.__totalTimeInSeconds // d) % m
                    for d, m in (
                        (604800, self.__maxWeeks),
                        (86400, 7),
                        (3600, 24),
                        (60, 60),
                        (1, 60),
                    )
                ],
                ["wk", "d", "hr", "min", "sec"],
            )
        ]

        # result.reverse()

        self.__seconds = result.pop()[0]
        self.__minutes = int(result.pop()[0])
        self.__hours = int(result.pop()[0])
        self.__days = int(result.pop()[0])
        self.__weeks = int(result.pop()[0])

    @property
    def weeks(self):
        return self.__weeks

    @property
    def days(self):
        return self.__days

    @property
    def hours(self):
        return self.__hours

    @property
    def minutes(self):
        return self.__minutes

    @property
    def seconds(self):
        return self.__seconds

    @property
    def total_seconds(self):
        return self.__totalTimeInSeconds


def time_span_tests():
    dt1 = datetime.datetime.now()
    dt2 = datetime.datetime.now() - datetime.timedelta(hours=1, minutes=5)

    ts = timespan(seconds=(dt1 - dt2).total_seconds())

    print("dt1: {}".format(dt1))
    print("dt2: {}".format(dt2))
    print(ts.to_dict())
    print(ts.to_str())

    ts.add_weeks(1)

    print(ts.to_dict())
    print(ts.to_str())


if __name__ == "__main__":
    time_span_tests()



#!/usr/bin/env python3
###################################################################################################

import jsonpickle
import json_fix

from plexapi.library import Library
from plexapi.collection import Collection
from plexapi.media import Media
from pmm_cfg_gen.utils.plex_utils import PlexItemHelper

###################################################################################################
class PlexLibraryCache:
    library: str
    collections: list[dict]
    items: list[dict]

    def __init__(self, libraryName: str) -> None:
        __libraryName = libraryName

        self.collections = list()
        self.items = list()

    def addCollection(self, collection: Collection):
        if not self.__isCollectionCached(collection):
            self.collections.append({"title": collection.title, "metadata": collection})

    def addMediaItem(self, item: Media):
        if not self.__isItemCached(item):
            itemTitle = PlexItemHelper.formatItemTitle(item) # type: ignore
            self.items.append({"title": itemTitle, "metadata": item})

    def to_cached_items_dict(self) -> dict:
        return {"collections": self.collections, "items": self.items}

    def to_stats_dict(self) -> dict:
        return {"collections": len(self.collections), "items": len(self.items)}

    def __json__(self):
        return self.toDict()

    def toDict(self) -> dict:
        return {
            "library": self.library,
            "items": self.to_cached_items_dict(),
            "stats": self.to_stats_dict(),
        }

    def __isCollectionCached(self, item) -> bool:
        it = next(
            (x for x in self.collections if x["title"] == item.title),
            None,
        )

        return it is not None

    def __isItemCached(self, item) -> bool:
        it = self.findByItem(item)

        return it is not None

    def findByItem(self, item):
        it = next(
            (
                x
                for x in self.items
                if x["title"] == PlexItemHelper.formatItemTitle(item) or 
                    ("orig_title" in x and x["orig_title"] == PlexItemHelper.formatItemTitle(item)) or 
                    ("alt_title" in x and x["alt_title"] == PlexItemHelper.formatItemTitle(item))
            ),
            None,
        )

        return it

class PlexCache:
    __cache: dict[str, PlexLibraryCache]

    def __init__(self) -> None:
        self.__cache = dict()

    def addLibraryEntry(self, library: Library) -> PlexLibraryCache:
        self.__cache[library.title1] = PlexLibraryCache(library.title1)

        return self.__cache[library.title1]

    def getLibraryCache(self, library):
        self.__cache[library.title1]

    def __json__(self):
        return self.toDict()

    def toDict(self):
        result = {
            "items": self.__cache,
            "stats": {"libraries": len(self.__cache.keys())},
        }

        return result



#!/usr/bin/env python3
###################################################################################################

import logging
import logging.config
import logging.handlers
import os

import coloredlogs
import yaml

###################################################################################################

class RollingFileHanderEx(logging.handlers.RotatingFileHandler):
    def __init__(
        self,
        filename,
        mode: str = "a",
        maxBytes: int = 0,
        backupCount: int = 0,
        encoding: str | None = None,
        delay: bool = False,
        errors: str | None = None,
    ) -> None:
        os.makedirs(os.path.dirname(filename), exist_ok=True)

        super().__init__(filename, mode, maxBytes, backupCount, encoding, delay, errors)

###################################################################################################

def setup_logging(
    default_path="logging.yaml", default_level=logging.INFO, env_key="LOG_CFG"
):
    """
    | **@author:** Prathyush SP
    | Logging Setup
    """
    path = default_path
    value = os.getenv(env_key, None)
    if value:
        path = value
    if os.path.exists(path):
        with open(path, "rt") as f:
            try:
                config = yaml.safe_load(f.read())
                logging.config.dictConfig(config)
            except Exception as e:
                print(e)
                print("Error in Logging Configuration. Using default configs")
                logging.basicConfig(level=default_level)
                coloredlogs.install(level=default_level)
    else:
        logging.basicConfig(level=default_level)
        coloredlogs.install(level=default_level)
        print("Failed to load configuration file. Using default configs")



#!/usr/bin/env python3
###################################################################################################

import logging

import jsonpickle.handlers
from plexapi.base import PlexObject, PlexPartialObject
from plexapi.library import LibrarySection
from plexapi.collection import Collection
from plexapi.video import Video, Movie, Show
from plexapi.audio import Artist
from plexapi.server import PlexServer
import re

from pmm_cfg_gen.utils.settings_utils_v1 import SettingsPlexLibrary, globalSettingsMgr

###################################################################################################

# @jsonpickle.handlers.register(PlexObject, base=True)
class PlexJsonHandler(jsonpickle.handlers.BaseHandler):
    def flatten(self, obj, data):
        """
         Flatten a Plex object into a dictionary. This is useful for generating JSON - RPC responses from server - side
         
         @param obj - The object to be flattened
         @param data - The dictionary to be flattened into. Should be an empty dictionary
         
         @return The dictionary with flattened objects as key / value pairs in the form of a dictionary ( field_name : value )
        """
        # Fix url raise
        # Set the _baseurl attribute of obj to the PlexServer.
        if isinstance(obj, PlexServer):
            setattr(obj, "_baseurl", globalSettingsMgr.settings.plex.serverUrl)
        # remove methods, private fields etc
        members = [
            attr
            for attr in dir(obj)
            if not callable(getattr(obj, attr))
            and not attr.startswith("__")
            and not attr.startswith("_")
        ]
        # flatten all members of the class and add them to data
        for normal_field in members:
            # we use context flatten - so its called handlers for given class
            data[normal_field] = self.context.flatten(getattr(obj, normal_field), {})
        return data

    def restore(self, obj):
        """
         Restore a previously saved object. This is called by : meth : ` save ` when the object is saved to the database.
         
         @param obj - The object to restore. This can be anything that can be converted to a database object by : meth : ` save `.
         
         @return The restored object. Note that this will be a copy of the object passed in but it is up to the caller to make sure that the object is not modified
        """
        return super().restore(obj)


jsonpickle.handlers.registry.register(PlexObject, PlexJsonHandler, True)

###################################################################################################

class PlexItemHelper:
    @classmethod
    def isPMMItem(cls, item: PlexPartialObject):
        """
         Checks if item is PMM item. This is used to detect items that are part of an emmy awards
         
         @param cls - class that contains the item
         @param item - item that is to be checked for PMM
         
         @return True if item is part of emmy awards False if it is not part of emmy a
        """
        # Check if the item is a PMM item.
        for label in item.labels:
            # Check if label is a PMM item
            if str(label.tag).strip() in [
                "Decade",
                "Emmy Awards",
                "Golden Globes Awards",
                "Top Actors",
                "Top Directors",
                "Oscars Winners Awards",
            ]:
                logging.getLogger("pmm_cfg_gen").debug(
                    "isPMMItem Found: {} - {}".format(item.title, label.tag)
                )
                return True

            # Check if item is a PMM item
            if str(item.title).strip() in [
                "Golden Globes Best Director Winners",
                "Golden Globes Best Picture Winners",
                "Oscars Best Director Winners",
                "Oscars Best Picture Winners",
                "Newly Released",
                "New Episodes",
                "TMDb Airing Today",
                "TMDb On The Air",
            ]:
                logging.getLogger("pmm_cfg_gen").debug(
                    "isPMMItem Found: {} - {}".format(item.title, label.tag)
                )
                return True

        return False
    
    @classmethod
    def cleanString(cls, title: str) -> str:
        """
         Cleans a title to make it suitable for display. This is a helper function for L { Title } and L { TitleFromText }.
         
         @param cls - The class that defines the type of title to be cleaned.
         @param s - The title to be cleaned. This should be a string that has no punctuation or other non - alphanumeric characters.
         
         @return A string that has all characters in the form of a - zA - Z0 - 9 and underscores
        """
        return title.replace("/", "-").replace("\\", "-").replace(":", "-").replace("*", "-").replace("?", "-").replace("\"", "-").replace("<", "-").replace(">", "-").replace("|", "-")

    @classmethod
    def formatString(cls, formatString: str, library : LibrarySection | None = None, collection : Collection | None = None, item : Video | Artist | None = None, pmm : dict | None = None, librarySettings : SettingsPlexLibrary | None = None, cleanTitleStrings : bool = False) -> str:
        """
        The format string to format. This is a string with placeholders to be substituted.
        
        @param cls - The class that owns the LibrarySection. Defaults to None.
        @param library - The LibrarySection to format. Defaults to None.
        @param collection - The Collection to format. Defaults to None.
        @param item - The Videobutton to format. Defaults to None.
        
        @return The formatted string if there are placeholders or the original string
        """

        result = formatString

        if librarySettings is not None:
            result = result.replace("{{library.path}}", librarySettings.path) if librarySettings.path else result
        
        if library is not None:
            result = result.replace("{{library.title}}", cls.cleanString(str(library.title)) if cleanTitleStrings else library.title)
            result = result.replace("{{library.type}}", library.type) if library.type else result
            # if no librarySettings.path defined lets use the library.title as a fall back
            result = result.replace("{{library.path}}", cls.cleanString(str(library.title)) if cleanTitleStrings else library.title)

        if collection is not None:
            result = result.replace("{{collection.title}}", cls.cleanString(collection.title) if cleanTitleStrings else collection.title)
            result = result.replace("{{collection.type}}", collection.type if collection.type else "")
            result = result.replace("{{collection.subtype}}", collection.subtype if collection.subtype else "")
            result = result.replace("{{collection.minYear}}", str(collection.minYear) if collection.minYear else "")
            result = result.replace("{{collection.minYear}}", str(collection.maxYear) if collection.maxYear else "")

            if "{{universe}}" in result:
                lstLabels = PlexItemHelper.getNamedCollectionLabels(collection)
                if lstLabels is not None and len(lstLabels) > 0:
                    logging.getLogger("pmm_cfg_gen").debug("formatString collection labels: {}".format(lstLabels))
                    result = result.replace("{{universe}}", lstLabels[0] if lstLabels[0] not in result else "")

        if item is not None:
            result = result.replace("{{item.title}}", cls.cleanString(item.title) if cleanTitleStrings else item.title)
            result = result.replace("{{item.titleSort}}", item.titleSort if item.titleSort else "")

            if isinstance(item, Video):
                result = result.replace("{{item.year}}", str(item.year) if item.year and str(item.year) not in item.title else "")
                result = result.replace("{{item.type}}", item.type if item.type else "")
                result = result.replace("{{item.contentRating}}", item.contentRating if item.contentRating else "")
                result = result.replace("{{item.editionTitle}}", item.editionTitle if isinstance(item, Movie) and item.editionTitle else "")
            elif isinstance(item, Artist):
                pass
    
            if "{{universe}}" in result:
                lstLabels = PlexItemHelper.getNamedCollectionLabels(item)
                if lstLabels is not None and len(lstLabels) > 0:
                    logging.getLogger("pmm_cfg_gen").debug("formatString item labels: {}".format(lstLabels))
                    result = result.replace("{{universe}}", lstLabels[0] if lstLabels[0] not in result else "")

        if pmm is not None:
            if "{{universe}}" in result:
                if "label" in pmm and len(pmm["label"]) > 0:
                    logging.getLogger("pmm_cfg_gen").debug("formatString pmm labels: {}".format(pmm))
                    result = result.replace("{{universe}}", pmm["label"][0] if pmm["label"][0] not in result else "")

        #logging.getLogger("pmm_cfg_gen").debug("formatString Before Cleaning: {} -> {}".format(formatString, result))

        # Lets remove any remaining tokens
        result = re.sub(r"\{\{[^\}]*\}\}", "", result)

        result = result.replace("()", "").replace("[]", "").strip()
        if result.startswith("-"):
            result = result[1:].strip()

        #logging.getLogger("pmm_cfg_gen").debug("formatString After Cleaning: {} -> {}".format(formatString, result))

        return result

    @classmethod
    def cleanItemTitle(cls, item : PlexPartialObject) -> str:
        """
         Clean the title of an item. This is a convenience method for L { cleanTitle }. It removes punctuation and other non - title characters that are illegal in Plex's title and returns the cleaned title.
         
         @param cls - The class to use for this method. Should be a subclass of L { PlexPartialObject }.
         @param item - The item to clean. Should be a L { PlexPartialObject }.
         
         @return The cleaned title of the item as a string. Will be empty if there is no title or it is not valid
        """
        return cls.cleanString(str(item.title))

    @classmethod
    def formatItemTitle(cls, item : Collection | Video | Artist, includeYear : bool = True, includeEdition : bool = True ) -> str:

        strFormat = "{{item.title}}"

        if isinstance(item, Collection):
            return PlexItemHelper.formatString(strFormat, collection=item)
        else:
            if isinstance(item, Video):
                if re.match(r"[\s\S]*\([\d]{4}\)$", item.title, flags=re.DOTALL) is None:
                    strFormat += " ({{item.year}})" if includeYear else ""
                strFormat += " [{{item.editionTitle}}]" if includeEdition and isinstance(item, Movie) else ""
            elif isinstance(item, Artist): 
                pass

            return PlexItemHelper.formatString(strFormat, item=item)

        return ""

    @classmethod
    def getItemLabels(cls, item: PlexPartialObject) -> list[str] | None:
        listResult = []

        listResult = [ x.tag for x in item.labels if x.tag != "PMM" ]

        return listResult

    @classmethod
    def getNamedCollectionLabels(cls, item : PlexPartialObject) -> list[str] | None:
        lstLabels = PlexItemHelper.getItemLabels(item)

        result = []
        if lstLabels is not None:
            result = [ x.replace("PMM-C-", "").replace("PMM-U-", "") for x in lstLabels if x.startswith("PMM-C-") or x.startswith("PMM-U-") ]
            
        return result 
    

class PlexCollectionHelper:
    __collection: Collection
    __guids: dict[str, list]

    def __init__(self, collection: Collection) -> None:
        """
         Initialize the instance. This is the method that must be called by the user to initialize the instance.
         
         @param collection - The collection of documents to operate on. If this is None the instance will be initialized with all documents in the collection
        """
        self.__collection = collection

        self.__process()

    def __process(self):
        """
         Process Guids @todo this needs to be refactored to use self. __parseGuids
        """
        self.__parseGuids()

    def __parseGuids(self):
        """
         Parse Guid's and store them in self. __guids @todo this needs to be
        """
        self.__guids = dict({"tmdb": list(), "tvdb": list(), "imdb": list()})

        # Add guids to the guids.
        for item in self.__collection.items():
            pih = PlexVideoHelper(item)

            # Add guids to the list of guids.
            for key in self.__guids.keys():
                self.__guids[key].append(pih.getGuidByName(key))

    def getGuidByName(self, name: str) -> list | None:
        """
         Get GUID by name. This is used to find the GUID that corresponds to a given name. If there is no such GUID None is returned
         
         @param name - Name of the GUID to look up
         
         @return List of GUIDs or None if not found ( in which case None is returned ) Note : The GUIDs are sorted by
        """
        return self.__guids[name] if name in self.__guids.keys() else None

    def getCollectionLabels(self) -> list[str] | None:
        return PlexItemHelper.getNamedCollectionLabels(self.__collection)
    
    @property
    def guids(self) -> dict[str, list]:
        """
         Returns the guids of the entity. This is a dictionary where keys are entity names and values are lists of GUIDs.
         
         
         @return the guids of the entity in the form of a dictionary where keys are entity names and values are lists of GUID
        """
        return self.__guids


class PlexVideoHelper:
    __item: PlexPartialObject
    __guids: dict[str, str]

    def __init__(self, item: PlexPartialObject):
        """
         Initialize the PlexPartialObject with data. This is called by __init__ and should not be called directly
         
         @param item - The item to be
        """
        self.__item = item
        self.__guids = {}

        self.__process()

    def __process(self):
        """
         Process Guids @todo this needs to be refactored to use self. __parseGuids
        """
        self.__parseGuids()

    def __parseGuids(self):
        """
         Parse guids and return a dictionary of guids. This is used to determine which items are part of the
        """
        try:
            # Set guids to a dict of guids
            if "guids" in self.__item.__dict__:
                self.__guids = dict(o.id.split("://") for o in self.__item.guids)
        except:
            pass

        try:
            # If the guid is not in the dict__ then the guid is replaced with the tvdb.
            if "guid" in self.__item.__dict__ and not self.__guids:
                # "guid": "com.plexapp.agents.thetvdb://73546?lang=en",
                try:
                    if self.__item.guid.startswith("com.plexapp.agents.thetvdb://"):
                        self.__guids["tvdb"] = self.__item.guid.replace(
                            "com.plexapp.agents.thetvdb://", ""
                        )

                    if self.__item.guid.startswith("com.plexapp.agents.themoviedb://"):
                        self.__guids["tmdb"] = self.__item.guid.replace(
                            "com.plexapp.agents.themoviedb://", ""
                        )

                    # If tvdb is not present in the guids tvdb then it will be removed from the guids tvdb.
                    if "?" in self.__guids["tvdb"]:
                        self.__guids["tvdb"] = str(self.__guids["tvdb"]).split("?")[0]

                    # If tvdb is not present in the guids tvdb then it will be removed from the guids tmdb.
                    if "?" in self.__guids["tmdb"]:
                        self.__guids["tmdb"] = str(self.__guids["tmdb"]).split("?")[0]
                except:
                    pass
        except:
            pass

        # Set the guid to the tmdb guid
        if "tmdb" not in self.__guids:
            self.__guids["tmdb"] = ""
        # Clear the guids for tvdb.
        if "tvdb" not in self.__guids:
            self.__guids["tvdb"] = ""
        # If imdb is not set this will remove the imdb guid from the guids
        if "imdb" not in self.__guids:
            self.__guids["imdb"] = ""

    def getGuidByName(self, name: str) -> str | None:
        """
         Get GUID by name. This is used to find the GUID that corresponds to a given entity name.
         
         @param name - Name of entity to look up. E. g.
         
         @return GUID or None if not found ( name not found ). Note that GUIDs are unique in the entity
        """
        return self.__guids[name] if name in self.__guids.keys() else None

    def getCollectionLabels(self) -> list[str] | None:
        return PlexItemHelper.getNamedCollectionLabels(self.__item)
    
    @property
    def guids(self) -> dict[str, str]:
        """
         Returns the guids of the entity. This is a dictionary with keys that correspond to the entity's GUID and values that correspond to the entity's name.
         
         
         @return the guids of the entity as a dictionary with keys that correspond to the entity's GUID and values that correspond
        """
        return self.__guids



#!/usr/bin/env python3
###################################################################################################

# import trakt
import logging
import jsonpickle

from pmm_cfg_gen.utils.settings_utils_v1 import globalSettingsMgr

###################################################################################################


# https://pytrakt.readthedocs.io/en/latest/getstarted.html
# https://github.com/moogar0880/PyTrakt
class TraktHelper:
    __logger: logging.Logger

    def __init__(self) -> None:
        self.__logger = logging.getLogger("pmm-cfg-gen")

        # self.__loggerFunc = self.__logger.debug
        self.__loggerFunc = print

        if (
            globalSettingsMgr.settings.theTvDatabase.apiKey is not None
            and len(globalSettingsMgr.settings.theTvDatabase.apiKey) > 0
            and globalSettingsMgr.settings.theTvDatabase.pin is not None
        ):
            self.__loggerFunc("Initalizing trakt connection")
        else:
            self.__tvdbApi = None

    def getListByName(self, name: str) -> str | None:
        self.__loggerFunc("Searching for trakt list: {}".format(name))

        # result = self.__tvdbApi.get_all_lists(meta= { "menu[type]" : "list", "query": name } )

        # self.__loggerFunc("trakt result: {}".format(jsonpickle.dumps(result, unpicklable=False)))

        return ""



#!/usr/bin/env python3
#######################################################################

import jsonpickle
import urllib.parse
import requests
import typing as t
import re
from jinja2.exceptions import FilterArgumentError

from dotty_dict import dotty

from plexapi.library import LibrarySection
from plexapi.collection import Collection
from plexapi.video import Video, Movie, Show

from pmm_cfg_gen.utils.settings_utils_v1 import globalSettingsMgr
from pmm_cfg_gen.utils.plex_utils import PlexItemHelper, PlexVideoHelper, PlexCollectionHelper
from pmm_cfg_gen.utils.tmdb_utils import TheMovieDatabaseHelper
# from pmm_cfg_gen.utils.tvdb_utils import TheTvDatabaseHelper

#######################################################################
# Jinja2 filters and utilitiy methods


def concat(*args: t.Any, **kwargs: t.Any) -> list:
    """
     Concatenates arguments to a list. This is useful for filters that need to combine iterables and lists into a single list
     
     @param args - The arguments to be concatenated
     @param kwargs - The keyword arguments to be concatenated
     
     @return A new list with all the arguments concatenated into it.
    """
    result = []

    # Add prefix postfix and prefix arguments to the arguments.
    if not args:    
        # Raise an exception if keyword arguments are not keyword arguments.
        if kwargs:
            raise FilterArgumentError(
                f"Unexpected keyword argument {next(iter(kwargs))!r}"
            )
    else:
        try:
            for a in args:
                if isinstance(a, (list, tuple)):
                    result.extend(a)
                else:
                    result.append(str(a))

        except LookupError:
            raise FilterArgumentError("concat requires a prefix and postfix argument") from None

    result = list(set(list(map(str, result))))

    while ("" in result):
        result.remove("")

    return result

def formatJson(data, indent: int = 4):
    """
     Formats data to JSON. This is a helper function for L { jsonpickle. dumps } that does not include unpicklable objects.
     
     @param data - The data to format. It must be a dictionary or an iterable
     @param indent - The number of spaces to indent the JSON.
     
     @return A JSON representation of the data as a string. The output is guaranteed to be UTF - 8 encoded
    """
    return jsonpickle.dumps(data, indent=indent, unpicklable=False)

def quote(data : str, quateChar : str = "\"") -> str:
    """
     Quote string
     
     @param data - The data to quote.
     @param quateChar - The character to use for quoting.
     
     @return The quoted data if it contains spaces otherwise the original data is returned as is. 
    """
    # Returns a string with the quate character.
    if " " in data.strip():
        return "{}{}{}".format(quateChar, data, quateChar)

    return data

def add_prepostfix(data : str, *args: t.Any, **kwargs: t.Any):
    """
     Add prefix and postfix to data if not already present. This is useful for filters that don't want to prefix or postfix the data to be sent to the filter
     
     @param data - The data to check for prefix and postfix
     @param args - The arguments to be checked for prefix and postfix
     @param kwargs - The keyword arguments to be checked for prefix and postfix
     
     @return The data with prefix and postfix added to it if not already present or stripped of any pre - or
    """
    prefix = None
    postfix = None 

    # Add prefix postfix and prefix arguments to the arguments.
    if not args:    
        # If prefix is not specified the prefix is used.
        if "prefix" in kwargs:
            prefix = kwargs.pop("prefix", None)

        # If postfix is not present in kwargs it will be removed from the postfix dictionary.
        if "postfix" in kwargs:
            postfix = kwargs.pop("postfix", None)

        # Raise an exception if keyword arguments are not keyword arguments.
        if kwargs:
            raise FilterArgumentError(
                f"Unexpected keyword argument {next(iter(kwargs))!r}"
            )
    else:
        try:
            prefix = args[0]
            postfix = args[1:]
        except LookupError:
            raise FilterArgumentError("add_prepostfix requires a prefix and postfix argument") from None

    # Append prefix to data if prefix is not empty
    if prefix and not data.strip().startswith(prefix):
        data = "{}{}".format(prefix, data)

    # Returns a string with the postfix appended to the data.
    if postfix and not data.strip().endswith(postfix):
        data = "{}{}".format(data, postfix)

    #print("add_prepostfix: {}".format(data))
    
    return data
        
def generateTpDbSearchUrl(item, baseUrl=None) -> str:
    """
     Generates URL forthePosterDatabase search. 
     
     @param item - The item to search for. It must have a title and subtype
     @param baseUrl - The base URL to use. If None the value from the PosterDatabase
     
     @return The URL that should be used to search for the item in Taxonomy Poster Database. If it's None the URL will be based on the settings
    """
    # Set the base URL to search URL if not set.
    if baseUrl is None:
        baseUrl = (
            globalSettingsMgr.settings.thePosterDatabase.searchUrlPro
            if globalSettingsMgr.settings.thePosterDatabase.enablePro
            else globalSettingsMgr.settings.thePosterDatabase.searchUrl
        )

    urlParms = {}

    urlParms.update({"term": item.title})

    if globalSettingsMgr.settings.thePosterDatabase.enablePro:
        if item.type == "collection":
            urlParms.update({"category": "Collections"})
        elif item.type == "movie":
            urlParms.update({"category": "Movies"})
            urlParms.update({"year_filter": "equals"})
            urlParms.update({"yone": item.year})
        elif item.type == "show":
            urlParms.update({"category": "Shows"})
            urlParms.update({"year_filter": "equals"})
            urlParms.update({"yone": item.year})
        elif item.type == "season":
            urlParms.update({"category": "Shows"})
            urlParms.update({"term": "{} {}".format(item.parentTitle, item.title)})

        if "guids" in item.__dict__:
            ids = dict(o.id.split("://") for o in item.guids)

            if "tmdb" in ids.keys():
                urlParms.update({"tmdb_id": ids["tmdb"]})
            if "imdb" in ids.keys():
                urlParms.update({"imdb_id": ids["imdb"]})
            if "tvdb" in ids.keys():
                urlParms.update({"tvdb_id": ids["tvdb"]})

    # logging.getLogger("pmm-cfg-gen").info(urlParms)

    urlQS = urllib.parse.urlencode(urlParms)

    req = requests.PreparedRequest()
    req.prepare_url(baseUrl, urlQS)

    return str(req.url)

def getItemGuidByName(item, guidName: str) -> str | None:
    """
     Get the Guid associated with an item. This is a wrapper around L { PlexItemHelper. getGuidByName }
     
     @param item - The item to get the guid from
     @param guidName - The name of the guid
     
     @return The Guid or None if not
    """
    s = ""

    if isinstance(item, Video):
        plexItem = PlexVideoHelper(item)

        s = plexItem.getGuidByName(guidName)
        #print("Video: Guid '{}' for item '{}' is '{}'".format(guidName, item.title, s))
    elif isinstance(item, Collection):
        plexCollection = PlexCollectionHelper(item)

        guids = plexCollection.getGuidByName(guidName)
        if guids: 
            s = ", ".join(guids)
        #print("Collection: Guid '{}' for item '{}' is '{}'".format(guidName.title, item, s))
    else: 
        s = ""

    return s

def getNamedCollectionLabels(item) -> list[str] | None:
    return PlexItemHelper.getNamedCollectionLabels(item)

def getCollectionGuidsByName(collection, guidName: str) -> list | None:
    """
     Get a list of Guid's from a PlexCollection. This is a convenience function to call L { PlexCollectionHelper }'s C { getGuidByName } method
     
     @param collection - A collection object or path to a collection
     @param guidName - The name of the Guid
     
     @return A list of G
    """
    pch = PlexCollectionHelper(collection)

    return pch.getGuidByName(guidName)

def getTmDbCollectionId(collection, tryExactMatch : bool = True) -> list[int] | None:
    """
     Get Tmdb collection ID. This is a wrapper around the TMDb findCollectionByName function to allow searching for collections by title
     
     @param collection - Collection to be looked up
     @param tryExactMatch - If True try to find exact match
     
     @return Collection ID or None if not found or no collection could be found in the TMDb ( not an exception
    """
    
    tmdbHelper = TheMovieDatabaseHelper()

    return tmdbHelper.findCollectionByName(collection.title, tryExactMatch)

def getPMMSeason(pmm : dict[str, dict], seasonNumber, attribute : str | None = None):
    """
     Get the poster URL for a season from the Plex Meta Manager season dictionary
     
     @param pmm - The Plex Meta Manager season dictionary
     @param seasonNumber - The season number
     
     @return The season or None if not found
    """
    if pmm is None: return None

    if "seasons" in pmm.keys():
        for season in pmm["seasons"]:
            if season == seasonNumber:
                return pmm["seasons"][season][attribute] if attribute in pmm["seasons"][season] else pmm["seasons"][season]

    return None

def getPMMAttributeByName(pmm : dict[str, dict], attribute : str | None = None):
    if pmm is None: return None

    tmp = dotty(pmm)

    if tmp.get(attribute) is not None: return tmp.get(attribute)
    if tmp.get("variables.{}".format(attribute)) is not None: return tmp.get("variables.{}".format(attribute))
    if tmp.get("template.{}".format(attribute)) is not None: return tmp.get("template.{}".format(attribute))

    return None

# def getTvDbListId(collection) -> list[int] | str | None:
#     tvDbHelper = TheTvDatabaseHelper()

#     return tvDbHelper.findListByName(collection.title)



#!/usr/bin/env python3
###################################################################################################

import logging
import os
from pathlib import Path
from typing import Any
import ruamel.yaml

###################################################################################################
class PlexMetaManagerCache:
    _logger: logging.Logger

    __collectionCache: dict
    __metadataCache: dict
    __overlayCache: dict

    def __init__(self) -> None:
        self._logger = logging.getLogger("pmm_cfg_gen")

        self.__collectionCache = {}
        self.__metadataCache = {}
        self.__overlayCache = {}

    def processFolder(self, path):
        self._logger.info("Processing Folder: '{}'".format(path)) 

        for root, dirs, files in os.walk(path):
                for file in files:
                    self.loadFile(Path(root, file)) 

        self._logger.info("Collections Processed: {}".format(len(self.__collectionCache)))
        self._logger.info("Metadata Processed: {}".format(len(self.__metadataCache)))
        self._logger.info("Overlays Processed: {}".format(len(self.__overlayCache)))
        
    def loadFile(self, fileName):
        self._logger.debug("Loading File: '{}'".format(fileName))

        with open(fileName, "r") as fp:
            try:
                data = ruamel.yaml.load(fp, Loader=ruamel.yaml.Loader)
                if data is None: return

                if "collections" in data:
                    for it in data["collections"]:
                        if it is not None:
                            self.__collectionCache[it] = data["collections"][it]

                if "metadata" in data:
                    for it in data["metadata"]:
                        if it is not None:
                            self.__metadataCache[it] = data["metadata"][it]

                if "overlays" in data:
                    for it in data["overlays"]:
                        if it is not None:
                            self.__overlayCache[it] = data["overlays"][it]

            except ruamel.yaml.YAMLError as exc:
                self._logger.error("Error parsing YAML file '{}'. Details: {}".format(fileName, exc))
            except Exception as exc:
                self._logger.error("Invalid PMM file '{}'. Details: {}".format(fileName, exc))

    def saveResultsToYaml(self, fileName : str | Path):
        self._logger.info("Saving results to file: '{}'".format(fileName))

        with open(fileName, "w") as fp:
            ruamel.yaml.round_trip_dump({"collections": self.__collectionCache, "metadata": self.__metadataCache}, fp)

    def collectionItem_to_dict(self, collectionName: str) -> dict[str, Any] | None:
        data = self.getCollectionCacheByName(collectionName)

        if data is None: return None

        self._logger.debug("collectionItem_to_dict: {}".format(data))
        
        result = {
            "title": data["title"] if "title" in data else collectionName,
            "label": self.__getAttributeListFromItemByName(data, "label"),
            "collection": self.__getAttributeListFromItemByName(data, "collection"),
            "list": self.__getAttributeListFromItemByName(data, "list"),
            "movie": self.__getAttributeListFromItemByName(data, "movie"),
            "show": self.__getAttributeListFromItemByName(data, "show"),
            "poster": self.__getPosterUrlFromItem(data),
            "trakt": self.__getAttributeListFromItemByName(data, "trakt_list"),
            "sort": {
                "prefix": self.__getAttributeFromItemByName(data, "sort_prefix"),
                "order": self.__getAttributeFromItemByName(data, "sort_order"),
                "separator": self.__getAttributeFromItemByName(data, "sort_separator"),
            }
        }

        return result

    def metadataItem_to_dict(self, metadataName: str, year : int | None = None) -> dict[str, Any] | None:
        data = self.getMetadataCacheByName(metadataName, year)

        if data is None: return None
        
        result = {
            "title": data["title"] if "title" in data else metadataName,

            "poster": self.__getPosterUrlFromItem(data),

            "label": self.__getAttributeListFromItemByName(data, "label"),

            "movie": self.__getAttributeListFromItemByName(data, "movie"),
            "show": self.__getAttributeListFromItemByName(data, "movie"),
            "list": self.__getAttributeListFromItemByName(data, "list"),
            "collection": self.__getAttributeListFromItemByName(data, "collection"),

            "sort_prefix": self.__getAttributeFromItemByName(data, "sort_prefix"),
            "sort_order": self.__getAttributeFromItemByName(data, "sort_order"),
            "sort_separator": self.__getAttributeFromItemByName(data, "sort_separator"),
        }

        if "seasons" in data:
            result.update( {"seasons": self.getPosterUrlsFromMetadataSeasons(metadataName, year) } )

        return result
    
    ###################################################################################################
    def getCollectionCacheByName(self, name : str) -> dict | None:
        self._logger.debug("Getting collection cache by name: '{}'".format(name))
        
        result = self.__collectionCache[name] if name in self.__collectionCache else None
        if result is None:
            result = next((x for x in self.__collectionCache if ("title" in x and x["title"] == name) or ("alt_title" in x and x["alt_title"] == name) or ("orig_title" in x and x["orig_title"] == name)), None)

        return result
    
    def getMetadataCacheByName(self, name : str, year : int | None) -> dict | None:
        self._logger.debug("Getting metadata cache by name: '{}'".format(name))
        
        result = self.__metadataCache[name] if name in self.__metadataCache else None

        if result is None:
            name = name.strip()
            self._logger.debug("Metadata cache by name: '{}' not found, searching by title".format(name))
            for k in self.__metadataCache:
                v = self.__metadataCache[k]
                #self._logger.info("Metadata cache by name: '{}' checking title: '{}'".format(name, x))
                if ("title" in v and v["title"].strip() == name) or ("alt_title" in v and v["alt_title"].strip() == name) or ("orig_title" in v and v["orig_title"].strip() == name): 
                    if year is None:
                        result = v 
                        
                        break
                    elif year-1 <= v["year"] <= year+1:
                        self._logger.debug("Metadata cache by name: '{}' found by title".format(name))
                        result = v

                        break
            
        return result

    def getOverlayCacheByName(self, name : str) -> dict | None:
        self._logger.debug("Getting overlay cache by name: '{}'".format(name))
        
        result = self.__overlayCache[name] if name in self.__overlayCache else None
        
        if result is None:
            result = next((x for x in self.__overlayCache if "title" in x and x["title"] == name), None)

        return result

    def getPosterUrlFromCollection(self, collectionName : str) -> str | None:
        data = self.getCollectionCacheByName(collectionName)
        if data is None:
            return None

        self._logger.debug("Getting Poster Url from collection: '{}'".format(collectionName))
        
        return self.__getPosterUrlFromItem(data)

    def getPosterUrlFromMetadata(self, metadataName : str, year : int | None) -> str | None:
        data = self.getMetadataCacheByName(metadataName, year)
        if data is None:
            return None

        self._logger.debug("Getting Poster Url from metadata: '{}'".format(metadataName))

        return self.__getPosterUrlFromItem(data)

    def getPosterUrlsFromMetadataSeasons(self, metadataName : str, year : int | None) -> dict[str, dict] | None:
        data = self.getMetadataCacheByName(metadataName, year)
        if data is None:
            return None

        self._logger.debug("Getting Poster Urls from metadata: '{}'".format(metadataName))

        result = {}
        if "seasons" in data and data["seasons"] is not None and len(data["seasons"]) > 0:
            for k in data["seasons"]:
                result[k] = { "poster": self.__getPosterUrlFromItem(data["seasons"][k]) }

        return result
    
    def getPosterUrlFromMetadataSeason(self, metadataName : str, season: int, year : int | None = None) -> str | None:
        result = self.getPosterUrlsFromMetadataSeasons(metadataName, year)

        if result is not None and str(year) in result.keys():
            return result[str(year)]["poster"] if "poster" in result[str(year)] else None
        
        return None

    def getShowListFromCollection(self, collectionName : str) -> list | None:
        data = self.getCollectionCacheByName(collectionName)
        if data is None:
            return None

        result = self.__getItemIDsFromCollectionByName(collectionName, "show")

        return result if type(result) is list else [result]

    def getMovieListFromCollection(self, collectionName : str) -> list | None:
        data = self.getCollectionCacheByName(collectionName)
        if data is None:
            return None

        result = self.__getItemIDsFromCollectionByName(collectionName, "movie")

        return result if type(result) is list else [result]

    def getCollectionListFromCollection(self, collectionName : str) -> list | None:
        data = self.getCollectionCacheByName(collectionName)
        if data is None:
            return None

        result = self.__getItemIDsFromCollectionByName(collectionName, "collection")

        return result if type(result) is list else [result]

    def getListsFromCollection(self, collectionName : str) -> list | None:
        data = self.getCollectionCacheByName(collectionName)
        if data is None:
            return None

        result = self.__getItemIDsFromCollectionByName(collectionName, "list")

        return result if type(result) is list else [result]

    ###################################################################################################
    def __getItemIDsFromCollectionByName(self, collectionName : str, listName : str) -> list | None:
        data = self.getCollectionCacheByName(collectionName)
        if data is None:
            return None
        
        self._logger.debug("Getting '{}' from collection: '{}'".format(listName, collectionName))
        self._logger.debug("Data: '{}'".format(data))

        result = self.__getAttributeListFromItemByName(data, listName)
            
        return result

    def __getAttributeListFromItemByName(self, data : dict, attribute : str) -> list | None:
        result = self.__stringCollectionToList(self.__getAttributeFromItemByName(data, attribute))

        return result

    def __getAttributeFromItemByName(self, data : dict, attribute : str) -> str | None:
        if data is None: return None
        
        self._logger.debug("Getting '{}' from item: '{}'".format(attribute, data))

        result = None

        if attribute in data: 
            self._logger.debug("Searching Root: '{}'".format(data[attribute]))
            return data[attribute]

        if "template" in data:
            self._logger.debug("Searching Template: '{}'".format(data["template"]))

            if isinstance(data["template"], list) and len(data["template"]) > 0:
                result = None
                for x in data["template"]:
                    if attribute in x: result =+ ", ".join(x[attribute])

                if result and len(result) > 0: return result.strip()
            
            if attribute in data["template"]: return data["template"][attribute]

        if "variables" in data:
            self._logger.debug("Searching Variable: '{}'".format(data["variables"]))
            if attribute in data["variables"]: return data["variables"][attribute]
            
        return result

    def __getPosterUrlFromItem(self, data : dict) -> str | None:
        if data is None: return None
        
        self._logger.debug("Getting Poster Url from: '{}'".format(data))
        
        posterUrl = self.__getAttributeListFromItemByName(data, "url_poster")
        if posterUrl is None or len(posterUrl) == 0: posterUrl = self.__getAttributeListFromItemByName(data, "poster")

        self._logger.debug("Poster Url: '{}'".format(posterUrl))
        
        if posterUrl is None or len(posterUrl) == 0: 
            return None

        return str(posterUrl[0]) if isinstance(posterUrl, list) else str(posterUrl)

    def __stringCollectionToList(self, collection : str | list | None) -> list[str]:
        if collection is None: return []        

        if isinstance(collection, list):
            return list(map(str, collection))
        
        return [x.strip() for x in collection.split(",")] if isinstance(collection, str) else [collection]
    
    ###################################################################################################
    ###################################################################################################
    ###################################################################################################
    ###################################################################################################
    
def test_PlexMetaManager(path: str, showName : str = "Bosch", year : int | None = None):
    pmm = PlexMetaManagerCache()

    p = Path(path)

    if p.exists():
        pmm.processFolder(p)

        pmm.saveResultsToYaml(Path(p.parent, Path(p.stem, "pmm_summary.yaml")))

        print("{} Poster: {}".format(showName, pmm.getPosterUrlFromCollection(showName)))
        print("{} Show List: {}".format(showName, pmm.getShowListFromCollection(showName)))
        print("{} Season 1 Poster: {}".format(showName, pmm.getPosterUrlFromMetadataSeason(showName, 1, 2015)))
    else:
        print("Path does not exist: {}".format(p))

if __name__ == "__main__":

    import sys
    logging.basicConfig(level=logging.INFO)
    test_PlexMetaManager(sys.argv[1], "Bosch", 2012) 


#!/usr/bin/env python3
###################################################################################################

import argparse
import logging

###################################################################################################

globalArgParser = argparse.ArgumentParser()

globalArgParser.add_argument(
    "--plex.serverUrl", 
    help="Plex Server fully qualifed URL"
)
globalArgParser.add_argument(
    "--plex.token", 
    help="Authentication Token (not claim token) for the plex server"
)
globalArgParser.add_argument(
    "--plex.lbraries", 
    nargs="*", 
    help="Comma delimited list of libraries to process"
)
globalArgParser.add_argument(
    "--output.path", 
    help="Root path to store generated files (default: ./data)"
)
globalArgParser.add_argument(
    "--output.overwrite",
    default=False,
    help="Overwrite existing files (default: %(default)s)",
)

globalArgParser.add_argument(
    "--theMovieDatabase.apiKey",
    help="The Movie Database API Key"
)
globalArgParser.add_argument(
    "--thePosterDatabase.enablePro",
    action="store_true",
    default=None,
    help="Enable Pro features for The Poster Database (requires you to be able to login to the site)"
)

globalArgParser.add_argument(
    "--pmm.deltaOnly",
    action="store_true",
    default=None,
    help="Only generate files for items that do not already exist"
)

# Advanced Arguments
globalArgParser.add_argument(
    "--generate.types",
    nargs="*",
    choices=["library.any", "collection.any", "collection.movie", "collection.show", "collection.music", "collection.report", "metadata.any", "metadata.movie", "metadata.show", "metadata.music", "metadata.report", "report.any", "overlay.any", "overlay.movie", "overlay.show", "overlay.music", "overlay.report"],
    #default=["collection.any", "metadata.any", "report.any", "overlay.any"],
    default=None,
    #help="Comma delimited list of item types to generate (default: %(default)s)",
    help=argparse.SUPPRESS
)
globalArgParser.add_argument(
    "--generate.formats",
    nargs="*",
    choices=["json", "yaml", "html"],
    #default=["json", "yaml", "html"],
    default=None,
    #help="Comma delimited list of formats to generate (default: %(default)s)",
    help=argparse.SUPPRESS
)

globalArgParser.add_argument(
    "--logLevel",
    choices=["INFO", "WARN", "DEBUG", "CRITICAL"],
    default="INFO",
    help="Logging Level (default: %(default)s)",
)

###################################################################################################

globalArgs = globalArgParser.parse_args()



# #!/usr/bin/env python3
# ###################################################################################################

# import tvdb_v4_official
# import logging
# import jsonpickle
# from pmm_cfg_gen.utils.settings_yml import globalSettingsMgr

# ###################################################################################################

# class TheTvDatabaseHelper:
#     __logger: logging.Logger
#     __tvdbApi : tvdb_v4_official.TVDB | None

#     def __init__(self) -> None:
#         self.__logger = logging.getLogger("pmm-cfg-gen")

#         #self.__loggerFunc = self.__logger.debug
#         self.__loggerFunc = print

#         if globalSettingsMgr.settings.theTvDatabase.apiKey is not None and len(globalSettingsMgr.settings.theTvDatabase.apiKey) > 0 and globalSettingsMgr.settings.theTvDatabase.pin is not None:
#             self.__loggerFunc("Initalizing tvdb connection")
#             self.__tvdbApi = tvdb_v4_official.TVDB(globalSettingsMgr.settings.theTvDatabase.apiKey, pin=globalSettingsMgr.settings.theTvDatabase.pin)
#         else:
#             self.__tvdbApi = None

#     def findListByName(self, name : str) -> list[int] | str:
#         if self.__tvdbApi is None: return ""

#         self.__loggerFunc("Searching for tvdb list: {}".format(name))

#         result = self.__tvdbApi.get_all_lists(meta= { "menu[type]" : "list", "query": name } )

#         self.__loggerFunc("tvdb result: {}".format(jsonpickle.dumps(result, unpicklable=False)))

#         return ""



#!/usr/bin/env python3
#######################################################################
# timer.py

import time
from contextlib import ContextDecorator
from dataclasses import dataclass, field
from typing import Any, Callable, ClassVar, Dict, Optional
from pmm_cfg_gen.utils.time_span import timespan

#######################################################################


class timer_exception(Exception):
    """A custom exception used to report errors in use of Timer class"""


@dataclass
class timer(ContextDecorator):
    """Time your code using a class, context manager, or decorator"""

    timers: ClassVar[Dict[str, float]] = {}
    name: Optional[str] = None
    text: str = "Elapsed time: {:0.4f} seconds"
    logger: Optional[Callable[[str], None]] = None

    _start_time: float = field(default=0, init=True, repr=False)
    _end_time: float = field(default=0, init=True, repr=False)
    _elapsed_time: timespan = field(default=Any, init=False, repr=False)
    _is_running: bool = field(default=False, init=True, repr=False)

    def __post_init__(self) -> None:
        """Initialization: add timer to dict of timers"""
        if self.name:
            self.timers.setdefault(self.name, 0)

    def start(self) -> None:
        """Start a new timer"""
        if self._is_running:
            raise timer_exception(f"Timer is running. Use .stop() to stop it")

        self._is_running = True
        self._start_time = time.perf_counter()

    def stop(self) -> None:
        """Stop the timer, and report the elapsed time"""
        if not self._is_running:
            raise timer_exception(f"Timer is not running. Use .start() to start it")

        # Calculate elapsed time
        self._end_time = time.perf_counter()
        self._elapsed_time = timespan(seconds=(self._end_time - self._start_time))
        self._is_running = False

        # Report elapsed time
        if self.logger:
            self.logger(self.text.format(self._elapsed_time))
        if self.name:
            self.timers[self.name] += self._elapsed_time.total_seconds

    @property
    def elapsed_time(self) -> float:
        return self._elapsed_time.total_seconds

    @property
    def elapsed_time_ts(self) -> timespan:
        return self._elapsed_time

    def __enter__(self) -> "timer":
        """Start a new timer as a context manager"""
        self.start()
        return self

    def __exit__(self, *exc_info: Any) -> None:
        """Stop the context manager timer"""
        self.stop()

    def to_dict(self):
        return {
            "start": self._start_time,
            "end": self._end_time,
            "elapsed": self._elapsed_time.to_dict(),
        }



#!/usr/bin/env python3
###################################################################################################

# from datetime import timedelta
# import time
from pmm_cfg_gen.utils.timer import timer

###################################################################################################

class PlexStatsTotals:
    total: int
    processed: int
    skipped: int

    def __init__(self) -> None:
        self.total = 0
        self.processed = 0
        self.skipped = 0

    def _addStats(self, stats):
        self.total += stats.total
        self.processed += stats.processed
        self.skipped += stats.skipped


class PlexStatsLibraryTotals:
    totals: PlexStatsTotals

    collections: PlexStatsTotals
    items: PlexStatsTotals

    def __init__(self) -> None:
        self.totals = PlexStatsTotals()
        self.collections = PlexStatsTotals()
        self.items = PlexStatsTotals()

    def calcTotals(self):
        self.totals.total = self.collections.total + self.items.total
        self.totals.processed = self.collections.processed + self.items.processed
        self.totals.skipped = self.collections.skipped + self.items.skipped


class PlexStats:
    timerProgram: timer
    timerLibraries: dict[str, timer]

    countsProgram: PlexStatsLibraryTotals
    countsLibraries: dict[str, PlexStatsLibraryTotals]

    def __init__(self) -> None:
        self.timerProgram = timer()
        self.timerLibraries = {}

        self.countsProgram = PlexStatsLibraryTotals()
        self.countsLibraries = {}

    def initLibrary(self, libraryName: str):
        self.timerLibraries[libraryName] = timer()
        self.countsLibraries[libraryName] = PlexStatsLibraryTotals()

    def calcTotals(self):
        # for libraryName in self.timerLibraries.keys():
        #     pass

        for libraryName in self.countsLibraries.keys():
            self.countsLibraries[libraryName].calcTotals()

            self.countsProgram.collections._addStats(
                self.countsLibraries[libraryName].collections
            )
            self.countsProgram.items._addStats(self.countsLibraries[libraryName].items)



#!/usr/bin/env python3

#######################################################################

import logging
import os
import readchar
import signal
import jsonpickle

from pmm_cfg_gen.utils.cli_args import globalArgs
from pmm_cfg_gen.utils.settings_utils_v1 import globalSettingsMgr
from pmm_cfg_gen.utils.logging_utils import setup_logging
from pmm_cfg_gen.utils.plex import PlexLibraryProcessor

#######################################################################

jsonpickle.set_preferred_backend("json")

#######################################################################


def handler(signum, frame):
    msg = "Ctrl-c was pressed. Do you really want to exit? y/n "
    print(msg, end="", flush=True)
    res = readchar.readchar()
    if res == "y":
        print("")
        os._exit(1)
    else:
        print("", end="\r", flush=True)
        print(" " * len(msg), end="", flush=True)  # clear the printed line
        print("    ", end="\r", flush=True)


signal.signal(signal.SIGINT, handler)

#######################################################################

setup_logging(
    str(globalSettingsMgr.modulePath.joinpath("logging.yaml")),
)

logger = logging.getLogger("pmm_cfg_gen")
logger.setLevel(getattr(logging, globalArgs.logLevel))

globalSettingsMgr.loadFromFile("config.yaml", globalArgs)

#######################################################################


def cli():
    plexMoveLibraryProcessor = PlexLibraryProcessor()
    plexMoveLibraryProcessor.process()


#######################################################################
#######################################################################

if __name__ == "__main__":
    cli()



#!/usr/bin/env python3
###################################################################################################

import themoviedb
import logging
import jsonpickle

from pmm_cfg_gen.utils.settings_utils_v1 import globalSettingsMgr

###################################################################################################


class TheMovieDatabaseHelper:
    __logger: logging.Logger
    __tmdbApi: themoviedb.TMDb | None

    def __init__(self) -> None:
        self.__logger = logging.getLogger("pmm-cfg-gen")

        self.__loggerFunc = self.__logger.debug
        # self.__loggerFunc = print

        if (
            globalSettingsMgr.settings.theMovieDatabase.apiKey is not None
            and len(globalSettingsMgr.settings.theMovieDatabase.apiKey) > 0
        ):
            self.__loggerFunc("Initalizing tmdb connection")
            self.__tmdbApi = themoviedb.TMDb(
                key=globalSettingsMgr.settings.theMovieDatabase.apiKey,
                language=globalSettingsMgr.settings.theMovieDatabase.language,
                region=globalSettingsMgr.settings.theMovieDatabase.region,
            )
        else:
            self.__tmdbApi = None

    def findCollectionByName(self, name: str, exactMatch: bool = False) -> list[int]:
        if self.__tmdbApi is None:
            return []

        name = name.strip()
        
        self.__loggerFunc("Searching for collection: '{}'".format(name))

        searchResults = self.__tmdbApi.search().collections(
            name
        ) 

        self.__loggerFunc(
            "tmdb result: {}".format(jsonpickle.dumps(searchResults, unpicklable=False))
        )

        results = []
        
        if searchResults is not None and searchResults.results is not None:
            if exactMatch:
                results = [x.id for x in searchResults.results if x.name == name or x.name == f"{name} Collection" ]

            if not exactMatch or len(results) == 0:
                results = [x.id for x in searchResults.results]

        if (
            globalSettingsMgr.settings.theMovieDatabase.limitCollectionResults is not None
            and globalSettingsMgr.settings.theMovieDatabase.limitCollectionResults > 0
            and results is not None and len(results) > 0
            and len(results) > globalSettingsMgr.settings.theMovieDatabase.limitCollectionResults
        ):
            results = results[
                : globalSettingsMgr.settings.theMovieDatabase.limitCollectionResults
            ]
        
        return results if results is not None else []



#!/usr/bin/env python3
#######################################################################

import logging
from pathlib import Path

import jinja2
import jinja2.exceptions
import jsonpickle

from pmm_cfg_gen.utils.settings_utils_v1 import globalSettingsMgr
from pmm_cfg_gen.utils.file_utils import writeFile
from pmm_cfg_gen.utils.plex_utils import PlexItemHelper, PlexVideoHelper, PlexCollectionHelper
import pmm_cfg_gen.utils.template_filters as template_filters

#######################################################################

class TemplateManager:
    __tplEnv: jinja2.Environment
    __cachedTemplates: dict

    #######################################################################
    def __init__(self, templatePath: str | Path) -> None:
        self._logger = logging.getLogger("pmm_cfg_gen")

        self._logger.debug(
            "Initializing Template Environment.  Template Path: '{}'".format(
                templatePath
            )
        )

        self.__tplEnv = jinja2.Environment(loader=jinja2.FileSystemLoader(templatePath))

        self.__cachedTemplates = {}
        self.__registerFilters()

    def render(self, templateName: str | Path, tplArgs: dict) -> str | None:
        self._logger.debug("Render data using template '{}'".format(templateName))

        tpl = self.__getTemplate(templateName)

        if tpl is None:
            self._logger.debug("Unable to load requested template: '{}'".format(templateName))

            return None

        if "settings" not in tplArgs.keys():
            tplArgs.update({"settings": globalSettingsMgr.settings})

        # self._logger.info("tplArgs: {}".format(tplArgs))
        return tpl.render(tplArgs)

    def renderAndSave(
        self, templateName: str | Path, fileName: str | Path, tplArgs: dict
    ):
        if templateName is None or templateName == "None":
            return

        self._logger.debug(
            "Render data using template '{}' and save to '{}'".format(
                templateName, fileName
            )
        )

        tplResult = self.render(templateName, tplArgs)

        if tplResult is not None:
            writeFile(fileName, tplResult)

    #######################################################################
    def __getTemplate(self, templateName: str | Path) -> jinja2.Template | None:
        if not templateName in self.__cachedTemplates.keys():
            self._logger.debug("Loading template into cache: {}".format(templateName))

            try:
                tpl = self.__tplEnv.get_template(str(templateName))

                self.__cachedTemplates[templateName] = tpl
            except jinja2.exceptions.TemplateNotFound:
                self._logger.debug("Requested Template does not exist: '{}'".format(templateName))

                return None
            except (
                jinja2.TemplateSyntaxError,
                jinja2.exceptions.UndefinedError,
            ) as exTpl:
                self._logger.error(
                    "Failed to load template: '{}'. Exception: {}".format(
                        templateName, str(exTpl)
                    )
                )
                if self._logger.isEnabledFor(logging.DEBUG):
                    self._logger.exception(
                        "Template Syntax Error: '{}'".format(templateName)
                    )

                return None
            except:
                if self._logger.isEnabledFor(logging.DEBUG):
                    self._logger.exception(
                        "Failed to load template: '{}'".format(templateName)
                    )

                return None

        else:
            self._logger.debug(
                "Retrieving template from cache: {}".format(templateName)
            )

        return self.__cachedTemplates[templateName]

    def __registerFilters(self):
        self.__tplEnv.filters["formatJson"] = template_filters.formatJson
        self.__tplEnv.filters["concat"] = template_filters.concat
        self.__tplEnv.filters["quote"] = template_filters.quote
        self.__tplEnv.filters["add_prepostfix"] = template_filters.add_prepostfix
        
        self.__tplEnv.filters["formatItemTitle"] = PlexItemHelper.formatItemTitle
        self.__tplEnv.filters["isPMMItem"] = PlexItemHelper.isPMMItem
        self.__tplEnv.filters["getPMMSeason"] = template_filters.getPMMSeason
        
        self.__tplEnv.filters["generateTpDbSearchUrl"] = template_filters.generateTpDbSearchUrl
        self.__tplEnv.filters["getItemGuidByName"] = template_filters.getItemGuidByName
        self.__tplEnv.filters["getNamedCollectionLabels"] = template_filters.getNamedCollectionLabels
        self.__tplEnv.filters["getCollectionGuidsByName"] = template_filters.getCollectionGuidsByName
        self.__tplEnv.filters["getTmDbCollectionId"] = template_filters.getTmDbCollectionId
        self.__tplEnv.filters["getPMMAttributeByName"] = template_filters.getPMMAttributeByName
        # self.__tplEnv.filters["getTvDbListId"] = template_filters.getTvDbListId



#!/usr/bin/env python3

#######################################################################

from typing import List
from enum import Enum

import logging
import os
from pathlib import Path

import confuse
import importlib_resources
import jsonpickle
from dotenv import load_dotenv
from expandvars import expandvars

#######################################################################

class SettingsTemplateFileFormatEnum(Enum):
    HTML = "html"
    JSON = "json"
    YAML = "yaml"


class SettingsTemplateLibraryTypeEnum(Enum):
    ANY = "any"
    MOVIE = "movie"
    MUSIC = "music"
    REPORT = "report"
    SHOW = "show"
    TEMPLATE = "template"


class SettingsGenerate:
    types: list[str]
    formats: list[str]

    def __init__(self, types: list[str], formats: list[str]) -> None:
        self.types = types
        self.formats = formats

    def isFormatEnabled(self, formatValue: SettingsTemplateFileFormatEnum | str) -> bool:
        if isinstance(formatValue, SettingsTemplateFileFormatEnum):
            formatValue = formatValue.value

        formatValue = str(formatValue).lower()

        logging.getLogger("pmm_cfg_gen").debug(f"Checking if format '{formatValue}' is enabled (formats: {self.formats})")

        return formatValue in self.formats
        
    def isTypeEnabled(self, typeValue: SettingsTemplateLibraryTypeEnum | str) -> bool:
        if isinstance(typeValue, SettingsTemplateLibraryTypeEnum):
            typeValue = typeValue.value

        typeValue = str(typeValue).lower()

        logging.getLogger("pmm_cfg_gen").debug(f"Checking if type '{typeValue}' is enabled (types: {self.types})")

        typeValueParts = typeValue.split(".")
        if len(typeValueParts) > 1:
            if f"{typeValueParts[0]}.any" in self.types:
                return True

        return typeValue in self.types


class SettingsOutputFileNames:
    collections: str
    metadata: str
    collectionsReport: str
    metadataReport: str
    report: str
    template: str

    def __init__(self, collections: str, metadata: str, collectionsReport: str, metadataReport: str, report: str, template: str) -> None:
        self.collections = collections
        self.metadata = metadata
        self.collectionsReport = collectionsReport
        self.metadataReport = metadataReport
        self.report = report
        self.template = template


class SettingsOutput:
    path: str
    pathFormat: str
    sharedTemplatePathFormat: str
    fileNameFormat: SettingsOutputFileNames
    overwrite: bool

    def __init__(self, path: str, pathFormat: str, sharedTemplatePathFormat: str, overwrite : bool, fileNameFormat: SettingsOutputFileNames) -> None:
        self.path = path
        self.pathFormat = pathFormat
        self.sharedTemplatePathFormat = sharedTemplatePathFormat
        self.fileNameFormat = fileNameFormat
        self.overwrite = overwrite


class SettingsPmmDefaults:
    deltaOnly: bool
    basePath: str | None

    def __init__(self, deltaOnly: bool = False, basePath: str | None = None) -> None:
        self.deltaOnly = deltaOnly
        self.basePath = expandvars(basePath.strip()) if basePath is not None else None        

class SettingsPlexLibrary:
    name: str
    path: str | None
    pmm_path: str | None
    pmm_delta: bool | None

    def __init__(self, name: str, path: str | None = None, pmm_path: str | None = None, pmm_delta: bool | None = None) -> None:
        self.name = name.strip()
        self.path = path.strip() if path is not None else name.strip()
        self.pmm_path = pmm_path.strip() if pmm_path is not None else None
        self.pmm_delta = pmm_delta 


class SettingsPlexServer:
    serverUrl: str
    token: str
    libraries: list[SettingsPlexLibrary]

    def __init__(self, serverUrl: str, token: str, libraries: List, pmmDefaults: SettingsPmmDefaults | None = None) -> None:
        self.serverUrl = serverUrl
        self.token = token

        if libraries is not None:
            self.libraries = []
            
            for library in libraries:
                if isinstance(library, dict):
                    self.libraries += [SettingsPlexLibrary(**library)]
                elif isinstance(library, str):
                    self.libraries += [SettingsPlexLibrary(name=library)]

        if pmmDefaults is not None:
            for library in self.libraries:
                
                if library.pmm_delta is None:
                    library.pmm_delta = pmmDefaults.deltaOnly
                    
                if library.path is None and pmmDefaults.basePath is not None:
                    library.path = os.path.join(pmmDefaults.basePath, library.name)
                        
    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            data["serverUrl"],
            data["token"],
            data["libraries"]
        )
                    

class SettingsPlexMetaManager:
    cacheExistingFiles: bool

    def __init__(self, cacheExistingFiles: bool) -> None: 
        self.cacheExistingFiles = cacheExistingFiles

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            data["cacheExistingFiles"],
            # [SettingsPlexMetaManagerFolder.from_dict(x) for x in data["folders"]]
        )


class SettingsTemplateFile:
    type: str
    format: str
    fileName: str
    fileExtension: str
    subFolder: str | None

    def __init__(self, type: str, format: str, fileName: str, fileExtension : str | None, subFolder : str | None) -> None:
        self.type = type
        self.format = format
        self.fileName = fileName
        self.fileExtension = fileExtension if fileExtension is not None else str(format).lower()
        self.subFolder = subFolder

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            data["type"],
            data["format"],
            data["file"],
            data["fileExtension"] if "fileExtension" in data else None,
            data["subFolder"] if "subFolder" in data else None,
        )

    @classmethod
    def from_list_dict(cls, data: List[dict]):
        return [cls.from_dict(x) for x in data]


class SettingsTemplateGroups:
    templatePath : str | None
    library: List[SettingsTemplateFile]
    collection: List[SettingsTemplateFile]
    metadata: List[SettingsTemplateFile]
    overlay: List[SettingsTemplateFile]

    def __init__(self, library: List[SettingsTemplateFile], collection: List[SettingsTemplateFile], metadata: List[SettingsTemplateFile], overlay: List[SettingsTemplateFile], templatePath : str | None) -> None:
        self.library = library
        self.collection = collection
        self.metadata = metadata
        self.overlay = overlay
        self.templatePath = templatePath

    def getTemplateRootPath(self) -> Path:
        if self.templatePath is None or self.templatePath == "pmm_cfg_gen.tempaltes":
            self.templatePath = str(
                importlib_resources.files("pmm_cfg_gen").joinpath("templates")
            )

        return Path(self.templatePath).resolve()

    def getTemplateByGroupName(self, name: str) -> List[SettingsTemplateFile]:

        result: List[SettingsTemplateFile] | None = None

        if name == "library":
            return self.library

        if name == "collection":
            return self.collection

        if name == "metadata":
            return self.metadata

        if name == "overlay":
            return self.overlay
        
        raise ValueError("Unknown template group name: '{}".format(name))

    def getTemplateByGroupAndLibraryType(self, group: str, libraryType: SettingsTemplateLibraryTypeEnum | str ) -> List[SettingsTemplateFile] | None:
        if isinstance(libraryType, SettingsTemplateLibraryTypeEnum):
            libraryType = str(libraryType.value)

        templateGroupList = self.getTemplateByGroupName(group)

        #logging.getLogger("pmm_cfg_gen").debug("templateGroupList: {}".format(jsonpickle.dumps(templateGroupList, unpicklable=False)))

        strGroupLibrary = f"{group}.{libraryType}".lower().strip()
        strGroupAny = f"{group}.any".lower().strip()

        #logging.getLogger("pmm_cfg_gen").debug("strGroupLibrary: '{}'".format(strGroupLibrary))
        #logging.getLogger("pmm_cfg_gen").debug("strGroupAny: '{}'".format(strGroupAny))
        
        result = [x for x in templateGroupList if x.type.lower() == strGroupLibrary or x.type.lower() == strGroupAny]

        if result is not None and len(result) > 0:
            return result 
        
        raise ValueError(
            "No template found for group '{}' and library type '{}'".format(
                group, libraryType
            )
        )


class SettingsTheTvDatabase:
    apiKey: str | None
    pin: str | None

    def __init__(self, apiKey: str | None, pin: str | None) -> None:
        self.apiKey = apiKey
        self.pin = pin


class SettingsTheMovieDatabase:
    limitCollectionResults: int
    apiKey: str
    language: str
    region: str

    def __init__(self, limitCollectionResults: int, apiKey: str, language: str, region: str) -> None:
        self.limitCollectionResults = limitCollectionResults
        self.apiKey = apiKey
        self.language = language
        self.region = region


class SettingsThePosterDatabase:
    enablePro: bool
    searchUrlPro: str
    searchUrl: str
    dbAssetUrl: str

    def __init__(self, enablePro: bool, searchUrlPro: str, searchUrl: str, dbAssetUrl: str) -> None:
        self.enablePro = enablePro
        self.searchUrlPro = searchUrlPro
        self.searchUrl = searchUrl
        self.dbAssetUrl = dbAssetUrl


class SettingsRunTime:
    currentWorkingPath: str
    currentWorkingPathRelative: str

    def __init__(self, currentWorkingPath: str) -> None:
        self.currentWorkingPath = currentWorkingPath

class Settings:
    version: str
    plex: SettingsPlexServer
    plexMetaManager: SettingsPlexMetaManager
    thePosterDatabase: SettingsThePosterDatabase
    theMovieDatabase: SettingsTheMovieDatabase
    theTvDatabase: SettingsTheTvDatabase
    templates: SettingsTemplateGroups
    output: SettingsOutput
    generate: SettingsGenerate
    runtime: SettingsRunTime

    def __init__(self, version: str, plex: SettingsPlexServer, plexMetaManager: SettingsPlexMetaManager, thePosterDatabase: SettingsThePosterDatabase, theMovieDatabase: SettingsTheMovieDatabase,  theTvDatabase : SettingsTheTvDatabase, templates: SettingsTemplateGroups, output: SettingsOutput, generate: SettingsGenerate, runtime: SettingsRunTime) -> None:
        self.version = version
        self.plex = plex
        self.plexMetaManager = plexMetaManager
        self.thePosterDatabase = thePosterDatabase
        self.theMovieDatabase = theMovieDatabase
        self.theTvDatabase = theTvDatabase
        self.templates = templates
        self.output = output
        self.generate = generate
        self.runtime = runtime

#######################################################################

class SettingsManager:
    _config: confuse.Configuration
    settings: Settings
    modulePath: Path

    def __init__(self) -> None:
        self._logger = logging.getLogger("pmm_cfg_gen")
        self.modulePath = Path(str(importlib_resources.files("pmm_cfg_gen")))

    def loadFromFile(self, fileName: str, cmdLineArgs=None) -> None:
        self._logger.debug("Loading Configuration")

        self._config = confuse.Configuration("pmm_cfg_gen", "pmm_cfg_gen")

        self._logger.debug("Loading Configuration from Default Configuration")
        self._config._add_default_source()
        self._config._add_user_source()

        self._config.set_file(self.modulePath.joinpath("config_default.yaml"))

        if os.path.exists(fileName):
            self._logger.debug("Loading Configuration from User Configuration: '{}'".format(fileName))
            self._config.set_file(fileName)

        self._logger.debug("Loading Configuration from Environment")
        load_dotenv(Path(os.getcwd(), ".env"))
        self._config.set_env("PMMCFG")

        if cmdLineArgs is not None:
            self._logger.debug("Loading Configuration from Command Line")
            self._logger.debug("Command Line Args: {}".format(cmdLineArgs))
            self._config.set_args(cmdLineArgs, dots=True)

        self._config.set_redaction("plex.token", True)

        self._logger.debug(
            "Configuration Directory: {}".format(self._config.config_dir())
        )
        self._logger.debug(
            "User Configuration Path: {}".format(self._config.user_config_path())
        )

        self._logger.debug("Loaded Configuration:")
        self._logger.debug(jsonpickle.dumps(self._config, unpicklable=False))

        self.settings = Settings(
            version=self._config["version"].get(confuse.Optional(str)),  # type: ignore
            plex=SettingsPlexServer(
                serverUrl=expandvars(self._config["plex"]["serverUrl"].as_str()),
                token=expandvars(self._config["plex"]["token"].as_str()),
                libraries=self._config["plex"]["libraries"].get(confuse.Optional(list)),  # type: ignore
                pmmDefaults=SettingsPmmDefaults(
                    deltaOnly=self._config["pmm"]["deltaOnly"].get(confuse.Optional(bool, default=None)),  # type: ignore
                    basePath=self._config["pmm"]["basePath"].get(confuse.Optional(str, default=None)),  # type: ignore
                )
            ),
            thePosterDatabase=SettingsThePosterDatabase(
                searchUrl=expandvars(
                    self._config["thePosterDatabase"]["searchUrl"].as_str()
                ),
                searchUrlPro=expandvars(
                    self._config["thePosterDatabase"]["searchUrlPro"].as_str()
                ),
                dbAssetUrl=expandvars(
                    self._config["thePosterDatabase"]["dbAssetUrl"].as_str()
                ),
                enablePro=bool(
                    self._config["thePosterDatabase"]["enablePro"].get(bool)
                ),
            ),
            theMovieDatabase=SettingsTheMovieDatabase(
                apiKey=self._config["theMovieDatabase"]["apiKey"].get(confuse.Optional(None)),  # type: ignore
                language=self._config["theMovieDatabase"]["language"].get(confuse.Optional(None)),  # type: ignore
                region=self._config["theMovieDatabase"]["region"].get(confuse.Optional(None)),  # type: ignore
                limitCollectionResults=self._config["theMovieDatabase"]["limitCollectionResults"].get(confuse.Optional(None)),  # type: ignore
            ),
            theTvDatabase=SettingsTheTvDatabase(
                apiKey=self._config["theTvDatabase"]["apiKey"].get(confuse.Optional(None)),  # type: ignore
                pin=self._config["theTvDatabase"]["pin"].get(confuse.Optional(None)),  # type: ignore
            ),
            templates=SettingsTemplateGroups(
                collection=SettingsTemplateFile.from_list_dict(self._config["templates"]["collection"].get(confuse.Optional(list))),  # type: ignore
                library=SettingsTemplateFile.from_list_dict(self._config["templates"]["library"].get(confuse.Optional(list))),  # type: ignore
                metadata=SettingsTemplateFile.from_list_dict(self._config["templates"]["metadata"].get(confuse.Optional(list))),  # type: ignore
                overlay=SettingsTemplateFile.from_list_dict(self._config["templates"]["overlay"].get(confuse.Optional(list))),  # type: ignore
                templatePath=self._config["templates"]["templatePath"].get(confuse.Optional(list)),  # type: ignore
            ),
            output=SettingsOutput(
                path=str(self._config["output"]["path"].as_str()),
                pathFormat=str(self._config["output"]["pathFormat"].as_str()),
                sharedTemplatePathFormat=str(self._config["output"]["sharedTemplatePathFormat"].as_str()),
                overwrite=bool(self._config["output"]["overwrite"].get(confuse.Optional(False))),
                fileNameFormat=SettingsOutputFileNames(
                    collections=str(
                        self._config["output"]["fileNameFormat"]["collections"].get(
                            confuse.Optional("{{collection.title}} ({{collection.year}})")
                        )
                    ),
                    metadata=str(
                        self._config["output"]["fileNameFormat"]["metadata"].get(
                            confuse.Optional("{{item.title}} ({{item.year}}) [{{item.editionTitle}}]")
                        )
                    ),
                    collectionsReport=str(
                        self._config["output"]["fileNameFormat"]["collectionsReport"].get(
                            confuse.Optional("{{library.title}} - Collections")
                        )
                    ),
                    metadataReport=str(
                        self._config["output"]["fileNameFormat"]["metedataReport"].get(
                            confuse.Optional("{{library.title}} - Items")
                        )
                    ),
                    report=str(self._config["output"]["fileNameFormat"]["report"].get(confuse.Optional("{{library.title}} -Report"))),
                    template=str(self._config["output"]["fileNameFormat"]["template"].get(confuse.Optional("template"))),
                )
            ),
            generate=SettingsGenerate(
                types=self._config["generate"]["types"].get(confuse.Optional(list)),  # type: ignore
                formats=self._config["generate"]["formats"].get(confuse.Optional(list)),  # type: ignore
            ),
            plexMetaManager=SettingsPlexMetaManager.from_dict(self._config["plexMetaManager"].get(confuse.Optional(dict))),  # type: ignore
            runtime=SettingsRunTime(
                currentWorkingPath=os.path.curdir
            ),
        )

        self._logger.debug("Active Settings:")
        self._logger.debug(jsonpickle.dumps(self.settings, unpicklable=False))


#######################################################################

globalSettingsMgr = SettingsManager()
